import logging

from smif.controller.build import build_model_run, get_model_run_definition
from smif.controller.load import load_resolution_sets
from smif.data_layer import DatafileInterface
from smif.modelrun import ModelRunError

LOGGER = logging.getLogger(__name__)


def execute_model_run(model_run_ids, directory, interface='local_binary', warm=False):
    """Runs the model run

    Parameters
    ----------
    modelrun_ids: list
        Modelrun ids that should be executed sequentially
    """
    LOGGER.info("Loading resolution data")
    load_resolution_sets(directory)

    model_run_definitions = []
    for model_run in model_run_ids:
        LOGGER.info("Getting model run definition for '" + model_run + "'")
        model_run_definitions.append(get_model_run_definition(directory, model_run))

    for model_run_config in model_run_definitions:

        LOGGER.info("Build model run from configuration data")
        modelrun = build_model_run(model_run_config)

        LOGGER.info("Running model run %s", modelrun.name)
        store = DatafileInterface(directory, interface)

        try:
            if warm:
                modelrun.run(store, store.prepare_warm_start(modelrun.name))
            else:
                modelrun.run(store)
        except ModelRunError as ex:
            LOGGER.exception(ex)
            exit(1)

        print("Model run '" + modelrun.name + "' complete")



from flask import jsonify, render_template
from smif.data_layer import (DataExistsError, DataMismatchError,
                             DataNotFoundError)
from smif.http_api.crud import (NarrativeAPI, NarrativeSetAPI, ScenarioAPI,
                                ScenarioSetAPI, SectorModelAPI, SmifAPI,
                                SosModelAPI, SosModelRunAPI)


def register_routes(app):
    """Register plain routing
    """
    @app.route('/')
    @app.route('/configure')
    @app.route('/configure/<path:path>')
    def home(path=None):
        """Render single page
        """
        return render_template('index.html')


def register_api_endpoints(app):
    """Register API calls (using pluggable views)
    """
    register_api(app, SmifAPI, 'smif_api', '/api/v1/smif/',
                 key='key', key_type='string')
    register_api(app, SosModelRunAPI, 'sos_model_run_api', '/api/v1/sos_model_runs/',
                 key='sos_model_run_name', key_type='string',
                 action='action', action_type='string')
    register_api(app, SosModelAPI, 'sos_model_api', '/api/v1/sos_models/',
                 key='sos_model_name', key_type='string')
    register_api(app, SectorModelAPI, 'sector_model_api', '/api/v1/sector_models/',
                 key='sector_model_name', key_type='string')
    register_api(app, ScenarioSetAPI, 'scenario_set_api', '/api/v1/scenario_sets/',
                 key='scenario_set_name', key_type='string')
    register_api(app, ScenarioAPI, 'scenario_api', '/api/v1/scenarios/',
                 key='scenario_name', key_type='string')
    register_api(app, NarrativeSetAPI, 'narrative_set_api', '/api/v1/narrative_sets/',
                 key='narrative_set_name', key_type='string')
    register_api(app, NarrativeAPI, 'narrative_api', '/api/v1/narratives/',
                 key='narrative_name', key_type='string')


def register_error_handlers(app):
    """Handle expected errors
    """
    @app.errorhandler(DataExistsError)
    def handle_exists(error):
        """Return 400 Bad Request if data to be created already exists
        """
        response = jsonify({"message": str(error)})
        response.status_code = 400
        return response

    @app.errorhandler(DataMismatchError)
    def handle_mismatch(error):
        """Return 400 Bad Request if data and id/name are mismatched
        """
        response = jsonify({"message": str(error)})
        response.status_code = 400
        return response

    @app.errorhandler(DataNotFoundError)
    def handle_not_found(error):
        """Return 404 if data is not found
        """
        response = jsonify({"message": str(error)})
        response.status_code = 404
        return response


def register_api(app, view, endpoint, url, key='id', key_type='int',
                 action=None, action_type=None):
    """Register a MethodView as an endpoint with CRUD operations at a URL
    """
    view_func = view.as_view(endpoint)
    app.add_url_rule(url, defaults={key: None},
                     view_func=view_func, methods=['GET'])
    app.add_url_rule(url, view_func=view_func, methods=['POST'])
    if action:
        app.add_url_rule('%s<%s:%s>/<%s:%s>' % (url, key_type, key, action_type, action),
                         view_func=view_func, methods=['GET'])
        app.add_url_rule('%s<%s:%s>/<%s:%s>' % (url, key_type, key, action_type, action),
                         view_func=view_func, methods=['POST'])
    app.add_url_rule('%s<%s:%s>' % (url, key_type, key), view_func=view_func,
                     methods=['GET', 'PUT', 'DELETE'])



# -*- coding: utf-8 -*-
"""Data access modules for loading system-of-systems model configuration
"""

# import classes for access like ::
#         from smif.data_layer import DatabaseInterface`
from smif.data_layer.database_interface import DatabaseInterface
from smif.data_layer.datafile_interface import DatafileInterface
from smif.data_layer.data_interface import (DataExistsError, DataMismatchError,
                                            DataNotFoundError)

from smif.data_layer.data_handle import DataHandle
from smif.data_layer.memory_interface import MemoryInterface

# Define what should be imported as * ::
#         from smif.data_layer import *
__all__ = ['DatabaseInterface', 'DatafileInterface', 'DataExistsError',
           'DataMismatchError', 'DataNotFoundError', 'DataHandle', 'MemoryInterface']



import os

from pytest import fixture
from smif.data_layer import DatafileInterface
from smif.data_layer.load import dump


@fixture(scope='function')
def get_handler_csv(setup_folder_structure, project_config):
    basefolder = setup_folder_structure
    project_config_path = os.path.join(
        str(basefolder), 'config', 'project.yml')
    dump(project_config, project_config_path)
    return DatafileInterface(str(basefolder), 'local_csv')


@fixture(scope='function')
def get_handler_binary(setup_folder_structure, project_config):
    basefolder = setup_folder_structure
    project_config_path = os.path.join(
        str(basefolder), 'config', 'project.yml')
    dump(project_config, project_config_path)
    return DatafileInterface(str(basefolder), 'local_binary')


@fixture(scope='function')
def get_remapped_scenario_data():
    """Return sample scenario_data
    """
    return [
        {
            'value': 100,
            'units': 'people',
            'region': 'oxford',
            'interval': 'cold_month',
            'year': 2015
        },
        {
            'value': 150,
            'units': 'people',
            'region': 'oxford',
            'interval': 'spring_month',
            'year': 2015
        },
        {
            'value': 200,
            'units': 'people',
            'region': 'oxford',
            'interval': 'hot_month',
            'year': 2015
        },
        {
            'value': 210,
            'units': 'people',
            'region': 'oxford',
            'interval': 'fall_month',
            'year': 2015
        },
        {
            'value': 100,
            'units': 'people',
            'region': 'oxford',
            'interval': 'cold_month',
            'year': 2016
        },
        {
            'value': 150,
            'units': 'people',
            'region': 'oxford',
            'interval': 'spring_month',
            'year': 2016
        },
        {
            'value': 200,
            'units': 'people',
            'region': 'oxford',
            'interval': 'hot_month',
            'year': 2016
        },
        {
            'value': 200,
            'units': 'people',
            'region': 'oxford',
            'interval': 'fall_month',
            'year': 2016
        }
    ]



from pytest import fixture, raises
from smif.decision import DecisionManager, PreSpecified, RuleBased


@fixture(scope='function')
def plan():
    planned_interventions = [
        {
         'name': 'small_pumping_station_oxford',
         'build_year': 2010
        },
        {
         'name': 'small_pumping_station_abingdon',
         'build_year': 2015

        },
        {
         'name': 'large_pumping_station_oxford',
         'build_year': 2020
        }
    ]

    return planned_interventions


@fixture(scope='function')
def get_strategies():
    strategies = [{'strategy': 'pre-specified-planning',
                   'description': 'build_nuclear',
                   'model_name': 'energy_supply',
                   'interventions': [{'name': 'nuclear_large', 'build_year': 2012},
                                     {'name': 'carrington_retire', 'build_year': 2011}]
                   }]

    return strategies


class TestPreSpecified:

    def test_initialisation(self, plan):

        timesteps = [2010, 2015, 2020]
        actual = PreSpecified(timesteps, plan)

        assert actual.timesteps == timesteps

    def test_generator(self, plan):

        timesteps = [2010, 2015, 2020]
        dm = PreSpecified(timesteps, plan)

        actual = next(dm)

        expected = {1: timesteps}

        assert actual == expected

    def test_get_state(self, plan):

        timesteps = [2010, 2015, 2020]
        dm = PreSpecified(timesteps, plan)

        actual = dm.get_state(2010)
        expected = {2010: ['small_pumping_station_oxford']}
        assert actual == expected

        actual = dm.get_state(2015)
        expected = {2010: ['small_pumping_station_oxford'],
                    2015: ['small_pumping_station_abingdon']}
        assert actual == expected

        actual = dm.get_state(2020)
        expected = {2010: ['small_pumping_station_oxford'],
                    2015: ['small_pumping_station_abingdon'],
                    2020: ['large_pumping_station_oxford']}
        assert actual == expected


class TestRuleBased:

    def test_initialisation(self):

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps)
        assert dm.timesteps == timesteps
        assert dm.satisfied is False
        assert dm.current_timestep_index == 0
        assert dm.current_iteration == 0

    def test_generator(self):

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps)

        actual = next(dm)
        assert actual == {1: [2010]}

        dm.satisfied = True
        actual = next(dm)
        assert actual == {2: [2015]}
        assert dm.satisfied is False

        actual = next(dm)
        assert actual == {3: [2015]}
        assert dm.satisfied is False

        dm.satisfied = True
        dm.current_timestep_index = 2
        actual = next(dm)
        assert actual is None


class TestDecisionManager():

    def test_null_strategy(self):
        strategy = []
        df = DecisionManager([2010, 2015], strategy)
        dm = df.decision_loop()
        bundle = next(dm)
        assert bundle == {0: [2010, 2015]}
        with raises(StopIteration):
            next(dm)

    def test_decision_manager_init(self, get_strategies):
        df = DecisionManager([2010, 2015], get_strategies)
        dm = df.decision_loop()
        bundle = next(dm)
        assert bundle == {0: [2010, 2015]}
        with raises(StopIteration):
            next(dm)

    def test_buildable(self, get_strategies):
        dm = PreSpecified([2010, 2015], get_strategies[0]['interventions'])
        assert dm.timesteps == [2010, 2015]
        assert dm.buildable(2010, 2010) is True
        assert dm.buildable(2011, 2010) is True

    def test_buildable_raises(self, get_strategies):
        dm = PreSpecified([2010, 2015], get_strategies[0]['interventions'])
        with raises(ValueError):
            dm.buildable(2015, 2014)

    def test_get_state(self, get_strategies):
        dm = PreSpecified([2010, 2015], get_strategies[0]['interventions'])
        actual = dm.get_state(2010)
        expected = {2011: ['carrington_retire'],
                    2012: ['nuclear_large']}
        assert actual == expected

        # actual = dm.get_state(2015)
        # expected = {2011: ['carrington_retire']}
        # assert actual == expected

        actual = dm.get_state(2015)
        expected = {2011: ['carrington_retire'],
                    2012: ['nuclear_large']}
        assert actual == expected



# -*- coding: utf-8 -*-
"""This module acts as a bridge to the sector models from the controller

The :class:`SectorModel` exposes several key methods for running wrapped
sector models.  To add a sector model to an instance of the framework,
first implement :class:`SectorModel`.

Utility Methods
===============
A number of utility methods are included to ease the integration of a
SectorModel wrapper within a System of Systems model.  These include:

- ``get_region_names(region_set_name)`` - gets a list of region names
- ``get_interval_names(interval_set_name)`` - gets a list of interval names

Key Functions
=============
This class performs several key functions which ease the integration of sector
models into the system-of-systems framework.

The user must implement the various abstract functions throughout the class to
provide an interface to the sector model, which can be called upon by the
framework. From the model's perspective, :class:`SectorModel` provides a bridge
from the sector-specific problem representation to the general representation
which allows reasoning across infrastructure systems.

The key functions include:

* converting input/outputs to/from geographies/temporal resolutions
* converting control vectors from the decision layer of the framework, to
  asset Interventions specific to the sector model
* returning scaler/vector values to the framework to enable measurements of
  performance, particularly for the purposes of optimisation and rule-based
  approaches

"""
import importlib
import logging
import os
from abc import ABCMeta, abstractmethod
from collections import defaultdict

from smif import StateData
from smif.convert.area import get_register as get_region_register
from smif.convert.interval import get_register as get_interval_register
from smif.model import Model

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


class SectorModel(Model, metaclass=ABCMeta):
    """A representation of the sector model with inputs and outputs

    Implement this class to enable integration of the wrapped simulation model
    into a system-of-system model.

    Arguments
    ---------
    name : str
        The unique name of the sector model

    Notes
    -----

    Implement the various abstract functions throughout the class to
    provide an interface to the simulation model, which can then be called
    upon by the framework.

    The key methods in the SectorModel class which need to be overridden are:

    - :py:meth:`SectorModel.initialise`
    - :py:meth:`SectorModel.simulate`
    - :py:meth:`SectorModel.extract_obj`

    A number of utility methods are included to ease the integration of a
    SectorModel wrapper within a System of Systems model.  These include:

    * ``get_region_names(region_set_name)`` - Get a list of region names
    * ``get_interval_names(interval_set_name)`` - Get a list of interval names

    For example, within the implementation of the simulate method, call::

        self.get_region_names('lad')

    to return a list of region names defined in the region register at runtime.

    """
    def __init__(self, name):
        super().__init__(name)

        self.path = ''
        self._initial_state = defaultdict(dict)
        self.interventions = []

        self.logger = logging.getLogger(__name__)

    def as_dict(self):
        """Serialize the SectorModel object as a dictionary

        Returns
        -------
        dict
        """
        config = {
            'name': self.name,
            'description': self.description,
            'path': self.path,
            'classname': self.__class__.__name__,
            'inputs': [inp.as_dict() for inp in self.inputs.values()],
            'outputs': [out.as_dict() for out in self.outputs.values()],
            'parameters': self.parameters.as_list(),
            'interventions': self.interventions,
            'initial_conditions': self._initial_state
        }
        return config

    def add_input(self, name, spatial_resolution, temporal_resolution, units):
        """Add an input to the sector model

        The inputs should be specified in a list.  For example::

                - name: electricity_price
                  spatial_resolution: GB
                  temporal_resolution: annual
                  units: £/kWh

        Arguments
        ---------
        name: str
        spatial_resolution: :class:`smif.convert.area.RegionSet`
        temporal_resolution: :class:`smif.convert.interval.IntervalSet`
        units: str

        """
        self.logger.debug("Adding input %s to %s", name, self.name)
        input_metadata = {"name": name,
                          "spatial_resolution": spatial_resolution,
                          "temporal_resolution": temporal_resolution,
                          "units": units}

        self.inputs.add_metadata(input_metadata)

    def add_output(self, name, spatial_resolution, temporal_resolution, units):
        """Add an output to the sector model

        Arguments
        ---------
        name: str
        spatial_resolution: :class:`smif.convert.area.RegionSet`
        temporal_resolution: :class:`smif.convert.interval.IntervalSet`
        units: str

        """
        self.logger.debug("Adding output %s to %s", name, self.name)
        output_metadata = {"name": name,
                           "spatial_resolution": spatial_resolution,
                           "temporal_resolution": temporal_resolution,
                           "units": units}

        self.outputs.add_metadata(output_metadata)

    def validate(self):
        """Validate that this SectorModel has been set up with sufficient data
        to run
        """
        pass

    @property
    def intervention_names(self):
        """The names of the interventions

        Returns
        =======
        list
            A list of the names of the interventions
        """
        return [intervention['name'] for intervention in self.interventions]

    @abstractmethod
    def initialise(self, initial_conditions):
        """Implement this method to set up the model system

        This method is called as the SectorModel is constructed, and prior to
        establishment of dependencies and other data links.

        Arguments
        ---------
        initial_conditions: list
            A list of past Interventions, with build dates and locations as
            necessary to specify the infrastructure system to be modelled.
        """
        pass

    def before_model_run(self, data):
        """Implement this method to conduct pre-model run tasks

        Arguments
        ---------
        data: smif.data_layer.DataHandle
            Access parameter values (before any model is run, no dependency
            input data or state is guaranteed to be available)
        """
        pass

    @abstractmethod
    def simulate(self, data):
        """Implement this method to run the model

        Arguments
        ---------
        data: smif.data_layer.DataHandle
            Access state, parameter values, dependency inputs

        Notes
        -----
        See docs on :class:`smif.data_layer.DataHandle` for details of how to
        access inputs, parameters and state and how to set results.

        ``interval``
            should reference an id from the interval set corresponding to
            the output parameter, as specified in model configuration
        ``region``
            should reference a region name from the region set corresponding to
            the output parameter, as specified in model configuration

        To obtain simulation model data in this method,
        use the data_handle methods such as::

            parameter_value = data.get_parameter('my_parameter_name')

        Other useful methods are ``get_base_timestep_data(input_name)``,
        ``get_previous_timestep_data(input_name)``,
        ``get_parameter(parameter_name)``, ``get_data(input_name, timestep=None)``,
        ``get_parameters()`` and
        ``get_results(output_name, model_name=None, modelset_iteration=None,
        decision_iteration=None, timestep=None)``.

        """
        pass

    @abstractmethod
    def extract_obj(self, results):
        """Implement this method to return a scalar value objective function

        This method should take the results from the output of the
        :py:meth:`simulate` method, process the results,
        and return a scalar value which can be used as a component of
        the objective function by the decision layer

        Arguments
        ---------
        results : dict
            A nested dict of the results from the :py:meth:`simulate` method

        Returns
        -------
        float
            A scalar component generated from the simulation model results
        """
        pass

    def get_region_names(self, region_set_name):
        """Get the unordered list of region names for
        ``region_set_name``

        Returns
        -------
        list
            A list of region names
        """
        return self.regions.get_entry(region_set_name).get_entry_names()

    def get_regions(self, region_set_name):
        """Get the list of regions for ``region_set_name``

        Returns
        -------
        list
            A list of GeoJSON-style dicts
        """
        return self.regions.get_entry(region_set_name).as_features()

    def get_region_centroids(self, region_set_name):
        """Get the list of region centroids for ``region_set_name``

        Returns
        -------
        list
            A list of GeoJSON-style dicts, with Point features corresponding to
            region centroids
        """
        return self.regions.get_entry(region_set_name).centroids_as_features()

    def get_interval_names(self, interval_set_name):
        """Get the list of interval names for ``interval_set_name``

        Returns
        -------
        list
            A list of interval names

        """
        return self.intervals.get_entry(interval_set_name).get_entry_names()


class SectorModelBuilder(object):
    """Build the components that make up a sectormodel from the configuration

    Parameters
    ----------
    name : str
        The name of the sector model
    sector_model : smif.model.SectorModel, default=None
        The sector model object

    Returns
    -------
    :class:`~smif.sector_model.SectorModel`

    Examples
    --------
    Call :py:meth:`SectorModelBuilder.construct` to populate a
    :class:`SectorModel` object and :py:meth:`SectorModelBuilder.finish`
    to return the validated and dependency-checked system-of-systems model.

    >>> builder = SectorModelBuilder(name, secctor_model)
    >>> builder.construct(config_data)
    >>> sos_model = builder.finish()

    """

    def __init__(self, name, sector_model=None):
        self._sector_model_name = name
        self._sector_model = sector_model
        self.interval_register = get_interval_register()
        self.region_register = get_region_register()
        self.logger = logging.getLogger(__name__)

    def construct(self, sector_model_config, timesteps):
        """Constructs the sector model

        Arguments
        ---------
        sector_model_config : dict
            The sector model configuration data
        """
        self.load_model(sector_model_config['path'], sector_model_config['classname'])
        self._sector_model.name = sector_model_config['name']
        self._sector_model.description = sector_model_config['description']
        self._sector_model.timesteps = timesteps

        self.add_inputs(sector_model_config['inputs'])
        self.add_outputs(sector_model_config['outputs'])
        self.add_interventions(sector_model_config['interventions'])
        self.create_initial_system(sector_model_config['initial_conditions'])
        self.add_initial_conditions(sector_model_config['initial_conditions'])
        self.add_parameters(sector_model_config['parameters'])

    def load_model(self, model_path, classname):
        """Dynamically load model module

        Arguments
        ---------
        model_path : str
            The path to the python module which contains the SectorModel
            implementation
        classname : str
            The name of the class of the SectorModel implementation

        """
        if os.path.exists(model_path):
            self.logger.info("Importing run module from %s", model_path)

            spec = importlib.util.spec_from_file_location(
                self._sector_model_name, model_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            klass = module.__dict__[classname]

            self._sector_model = klass(self._sector_model_name)
            self._sector_model.name = self._sector_model_name
            self._sector_model.path = model_path

        else:
            msg = "Cannot find '{}' for the '{}' model".format(
                model_path, self._sector_model_name)
            raise FileNotFoundError(msg)

    def add_initial_conditions(self, initial_conditions):
        """Adds initial conditions (state) for a model
        """
        state_data = filter(
            lambda d: len(d.data) > 0,
            [self.intervention_state_from_data(datum) for datum in initial_conditions]
        )
        self._sector_model._initial_state = list(state_data)

    def create_initial_system(self, initial_conditions):
        """Set up model with initial system
        """
        msg = "Sector model must be loaded before creating initial system"
        assert self._sector_model is not None, msg

        self._sector_model.initialise(initial_conditions)

    def add_parameters(self, parameter_config):
        """Add parameter configuration to sector model

        Arguments
        ---------
        parameter_config : list
            A list of dicts with keys ``name``, ``description``,
            ``absolute_range``, ``suggested_range``, ``default_value``,
            ``units``, ``parent``
        """

        for parameter in parameter_config:
            self._sector_model.add_parameter(parameter)

    def add_inputs(self, input_dicts):
        """Add inputs to the sector model
        """
        msg = "Sector model must be loaded before adding inputs"
        assert self._sector_model is not None, msg

        if input_dicts:

            for model_input in input_dicts:
                name = model_input['name']

                spatial_resolution = model_input['spatial_resolution']
                region_set = self.region_register.get_entry(spatial_resolution)

                temporal_resolution = model_input['temporal_resolution']
                interval_set = self.interval_register.get_entry(temporal_resolution)

                units = model_input['units']

                self._sector_model.add_input(name,
                                             region_set,
                                             interval_set,
                                             units)

    def add_outputs(self, output_dicts):
        """Add outputs to the sector model
        """
        msg = "Sector model must be loaded before adding outputs"
        assert self._sector_model is not None, msg

        if output_dicts:

            for model_output in output_dicts:
                name = model_output['name']

                spatial_resolution = model_output['spatial_resolution']
                region_set = self.region_register.get_entry(spatial_resolution)

                temporal_resolution = model_output['temporal_resolution']
                interval_set = self.interval_register.get_entry(temporal_resolution)

                units = model_output['units']

                self._sector_model.add_output(name,
                                              region_set,
                                              interval_set,
                                              units)

    def add_interventions(self, intervention_list):
        """Add interventions to the sector model

        Parameters
        ----------
        intervention_list : list
            A list of dicts of interventions

        """
        msg = "Sector model must be loaded before adding interventions"
        assert self._sector_model is not None, msg

        self._sector_model.interventions = intervention_list

    @staticmethod
    def intervention_state_from_data(intervention_data):
        """Unpack an intervention from the initial system to extract StateData
        """
        target = None
        data = {}
        for key, value in intervention_data.items():
            if key == "name":
                target = value

            if isinstance(value, dict) and "is_state" in value and value["is_state"]:
                del value["is_state"]
                data[key] = value

        return StateData(target, data)

    def validate(self):
        """Check and/or assert that the sector model is correctly set up
        - should raise errors if invalid
        """
        assert self._sector_model is not None, "Sector model not loaded"
        self._sector_model.validate()
        return True

    def finish(self):
        """Validate and return the sector model
        """
        self.validate()
        return self._sector_model



import numpy as np
from pytest import fixture


@fixture(scope='module')
def year_to_month_coefficients():
    """From one year to 12 months

    (apportions)
    """

    month_lengths = np.array([[31, 28, 31, 30, 31, 31, 30, 30, 31, 31, 30, 31]],
                             dtype=np.float).T
    return month_lengths / 365


@fixture(scope='module')
def month_to_year_coefficients():
    """
    """
    return np.ones((1, 12), dtype=np.float)


@fixture(scope='module')
def month_to_season_coefficients():
    """
    12 months to four seasons (winter is December, January, Feb)

    Sum value for each month into season
    """
    coef = np.array([
        [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],  # winter
        [0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],  # spring
        [0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0],  # summer
        [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0]])  # autumn

    return coef.T


@fixture(scope='module')
def season_to_month_coefficients():
    """
    12 months to four seasons (winter is December, January, Feb)

    To convert from seasons to months, find the proportion of each season that
    corresponds to the relevant month.

    E.g. winter to january is (duration of Jan / total duration of winter)
    """
    coef = np.array(
        # winter
        #     spring
        #        summer
        #           autumn
        [[31, 0, 0, 0],  # January
         [28, 0, 0, 0],  # Feb
         [0, 31, 0, 0],  # March
         [0, 30, 0, 0],  # April
         [0, 31, 0, 0],  # May
         [0, 0, 30, 0],  # June
         [0, 0, 31, 0],  # July
         [0, 0, 31, 0],  # August
         [0, 0, 0, 30],  # September
         [0, 0, 0, 31],  # October
         [0, 0, 0, 30],  # November
         [31, 0, 0, 0]]   # December
    )

    days_in_seasons = np.array([
        31+31+28,  # winter
        31+30+31,  # spring
        30+31+31,  # summer
        30+31+30  # autumn
    ], dtype=float)

    return np.transpose(coef / days_in_seasons)


@fixture(scope='function')
def months_csv():
    data = [
        {'id': 'jan', 'start': 'P0M', 'end': 'P1M'},
        {'id': 'feb', 'start': 'P1M', 'end': 'P2M'},
        {'id': 'mar', 'start': 'P2M', 'end': 'P3M'},
        {'id': 'apr', 'start': 'P3M', 'end': 'P4M'},
        {'id': 'may', 'start': 'P4M', 'end': 'P5M'},
        {'id': 'jun', 'start': 'P5M', 'end': 'P6M'},
        {'id': 'jul', 'start': 'P6M', 'end': 'P7M'},
        {'id': 'aug', 'start': 'P7M', 'end': 'P8M'},
        {'id': 'sep', 'start': 'P8M', 'end': 'P9M'},
        {'id': 'oct', 'start': 'P9M', 'end': 'P10M'},
        {'id': 'nov', 'start': 'P10M', 'end': 'P11M'},
        {'id': 'dec', 'start': 'P11M', 'end': 'P12M'}
    ]

    return data


@fixture(scope='function')
def months():
    data = [
        ('jan', [('P0M', 'P1M')]),
        ('feb', [('P1M', 'P2M')]),
        ('mar', [('P2M', 'P3M')]),
        ('apr', [('P3M', 'P4M')]),
        ('may', [('P4M', 'P5M')]),
        ('jun', [('P5M', 'P6M')]),
        ('jul', [('P6M', 'P7M')]),
        ('aug', [('P7M', 'P8M')]),
        ('sep', [('P8M', 'P9M')]),
        ('oct', [('P9M', 'P10M')]),
        ('nov', [('P10M', 'P11M')]),
        ('dec', [('P11M', 'P12M')])
    ]

    return data


@fixture(scope='function')
def remap_months_csv():
    """Remapping four representative months to months across the year

    In this case we have a model which represents the seasons through
    the year using one month for each season. We then map the four
    model seasons 1, 2, 3 & 4 onto the months throughout the year that
    they represent.

    The data will be presented to the model using the four time intervals,
    1, 2, 3 & 4. When converting to hours, the data will be replicated over
    the year.  When converting from hours to the model time intervals,
    data will be averaged and aggregated.

    """
    data = [{'id': 'cold_month', 'start': 'P0M', 'end': 'P1M'},
            {'id': 'cold_month', 'start': 'P1M', 'end': 'P2M'},
            {'id': 'spring_month', 'start': 'P2M', 'end': 'P3M'},
            {'id': 'spring_month', 'start': 'P3M', 'end': 'P4M'},
            {'id': 'spring_month', 'start': 'P4M', 'end': 'P5M'},
            {'id': 'hot_month', 'start': 'P5M', 'end': 'P6M'},
            {'id': 'hot_month', 'start': 'P6M', 'end': 'P7M'},
            {'id': 'hot_month', 'start': 'P7M', 'end': 'P8M'},
            {'id': 'fall_month', 'start': 'P8M', 'end': 'P9M'},
            {'id': 'fall_month', 'start': 'P9M', 'end': 'P10M'},
            {'id': 'fall_month', 'start': 'P10M', 'end': 'P11M'},
            {'id': 'cold_month', 'start': 'P11M', 'end': 'P12M'}]
    return data


@fixture(scope='function')
def remap_months():
    data = [
        ('cold_month', [('P0M', 'P1M'), ('P1M', 'P2M'), ('P11M', 'P12M')]),
        ('spring_month', [('P2M', 'P3M'), ('P3M', 'P4M'), ('P4M', 'P5M')]),
        ('hot_month', [('P5M', 'P6M'), ('P6M', 'P7M'), ('P7M', 'P8M')]),
        ('fall_month', [('P8M', 'P9M'), ('P9M', 'P10M'), ('P10M', 'P11M')])
        ]
    return data


@fixture(scope='function')
def seasons_csv():
    # NB "winter" is split into two pieces around the year end
    data = [{'id': 'winter', 'start': 'P0M', 'end': 'P2M'},
            {'id': 'spring', 'start': 'P2M', 'end': 'P5M'},
            {'id': 'summer', 'start': 'P5M', 'end': 'P8M'},
            {'id': 'autumn', 'start': 'P8M', 'end': 'P11M'},
            {'id': 'winter', 'start': 'P11M', 'end': 'P12M'}]
    return data


@fixture(scope='function')
def seasons():
    # NB "winter" is split into two pieces around the year end
    data = [('winter', [('P0M', 'P2M'), ('P11M', 'P12M')]),
            ('spring', [('P2M', 'P5M')]),
            ('summer', [('P5M', 'P8M')]),
            ('autumn', [('P8M', 'P11M')])]
    return data


@fixture(scope='function')
def twenty_four_hours_csv():
    data = [
        {'id': '1_0', 'start': 'PT0H', 'end': 'PT1H'},
        {'id': '1_1', 'start': 'PT1H', 'end': 'PT2H'},
        {'id': '1_2', 'start': 'PT2H', 'end': 'PT3H'},
        {'id': '1_3', 'start': 'PT3H', 'end': 'PT4H'},
        {'id': '1_4', 'start': 'PT4H', 'end': 'PT5H'},
        {'id': '1_5', 'start': 'PT5H', 'end': 'PT6H'},
        {'id': '1_6', 'start': 'PT6H', 'end': 'PT7H'},
        {'id': '1_7', 'start': 'PT7H', 'end': 'PT8H'},
        {'id': '1_8', 'start': 'PT8H', 'end': 'PT9H'},
        {'id': '1_9', 'start': 'PT9H', 'end': 'PT10H'},
        {'id': '1_10', 'start': 'PT10H', 'end': 'PT11H'},
        {'id': '1_11', 'start': 'PT11H', 'end': 'PT12H'},
        {'id': '1_12', 'start': 'PT12H', 'end': 'PT13H'},
        {'id': '1_13', 'start': 'PT13H', 'end': 'PT14H'},
        {'id': '1_14', 'start': 'PT14H', 'end': 'PT15H'},
        {'id': '1_15', 'start': 'PT15H', 'end': 'PT16H'},
        {'id': '1_16', 'start': 'PT16H', 'end': 'PT17H'},
        {'id': '1_17', 'start': 'PT17H', 'end': 'PT18H'},
        {'id': '1_18', 'start': 'PT18H', 'end': 'PT19H'},
        {'id': '1_19', 'start': 'PT19H', 'end': 'PT20H'},
        {'id': '1_20', 'start': 'PT20H', 'end': 'PT21H'},
        {'id': '1_21', 'start': 'PT21H', 'end': 'PT22H'},
        {'id': '1_22', 'start': 'PT22H', 'end': 'PT23H'},
        {'id': '1_23', 'start': 'PT23H', 'end': 'PT24H'}
    ]
    return data


@fixture(scope='function')
def twenty_four_hours():

    data = [
            ('1_0', [('PT0H', 'PT1H')]),
            ('1_1', [('PT1H', 'PT2H')]),
            ('1_2', [('PT2H', 'PT3H')]),
            ('1_3', [('PT3H', 'PT4H')]),
            ('1_4', [('PT4H', 'PT5H')]),
            ('1_5', [('PT5H', 'PT6H')]),
            ('1_6', [('PT6H', 'PT7H')]),
            ('1_7', [('PT7H', 'PT8H')]),
            ('1_8', [('PT8H', 'PT9H')]),
            ('1_9', [('PT9H', 'PT10H')]),
            ('1_10', [('PT10H', 'PT11H')]),
            ('1_11', [('PT11H', 'PT12H')]),
            ('1_12', [('PT12H', 'PT13H')]),
            ('1_13', [('PT13H', 'PT14H')]),
            ('1_14', [('PT14H', 'PT15H')]),
            ('1_15', [('PT15H', 'PT16H')]),
            ('1_16', [('PT16H', 'PT17H')]),
            ('1_17', [('PT17H', 'PT18H')]),
            ('1_18', [('PT18H', 'PT19H')]),
            ('1_19', [('PT19H', 'PT20H')]),
            ('1_20', [('PT20H', 'PT21H')]),
            ('1_21', [('PT21H', 'PT22H')]),
            ('1_22', [('PT22H', 'PT23H')]),
            ('1_23', [('PT23H', 'PT24H')])]
    return data


@fixture(scope='function')
def one_day_csv():
    data = [{'id': 'one_day', 'start': 'P0D', 'end': 'P1D'}]
    return data


@fixture(scope='function')
def one_day():
    data = [('one_day', [('P0D', 'P1D')])]
    return data


@fixture(scope='function')
def one_year_csv():
    data = [{'id': 'one_year', 'start': 'P0Y', 'end': 'P1Y'}]
    return data


@fixture(scope='function')
def one_year():
    data = [('one_year', [('P0Y', 'P1Y')])]
    return data


@fixture(scope='function')
def monthly_data():
    """[31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    """
    data = np.array([[
        31,
        28,
        31,
        30,
        31,
        30,
        31,
        31,
        30,
        31,
        30,
        31,
    ]])
    return data


@fixture(scope='function')
def monthly_data_as_seasons():
    return np.array([[
        31+31+28,
        31+30+31,
        30+31+31,
        30+31+30
    ]], dtype=float)


@fixture(scope='function')
def remap_month_data():
    data = np.array([[
        31+31+28,  # Dec, Jan, Feb
        31+30+31,  # Mar, Apr, May
        30+31+31,  # Jun, Jul, Aug
        30+31+30  # Sep, Oct, Nov
    ]], dtype=float) / 3

    return data


@fixture(scope='function')
def remap_month_data_as_months():
    data = np.array([[
        30.666666666,
        29.666666666,
        29.666666666,
        29.666666666,
        30.666666666,
        30.666666666,
        30.666666666,
        30.666666666,
        30.666666666,
        30.666666666,
        30.666666666,
        30.666666666
    ]])
    return data



"""Scenario models represent scenario data sources within a system-of-systems
model.
"""
from logging import getLogger

from smif.model import Model


class ScenarioModel(Model):
    """Represents exogenous scenario data

    Arguments
    ---------
    name : str
        The unique name of this scenario

    Attributes
    ----------
    name : str
        Name of this scenario
    timesteps : list
        List of timesteps for which the scenario holds data
    scenario_set : str
        Scenario set to which this scenario belongs
    """

    def __init__(self, name):
        super().__init__(name)
        self.scenario_set = None
        self.scenario_name = None

    def as_dict(self):
        config = {
            'name': self.name,
            'description': self.description,
            'scenario_set': self.scenario_set,
            'facets': [
                output.as_dict()
                for output in self.outputs.values()
            ]
        }
        return config

    def add_output(self, name, spatial_resolution, temporal_resolution, units):
        """Add an output to the scenario model

        Arguments
        ---------
        name: str
        spatial_resolution: :class:`smif.convert.area.RegionRegister`
        temporal_resolution: :class:`smif.convert.interval.TimeIntervalRegister`
        units: str

        """
        output_metadata = {
            "name": name,
            "spatial_resolution": spatial_resolution,
            "temporal_resolution": temporal_resolution,
            "units": units
        }
        self.outputs.add_metadata(output_metadata)

    def _check_output(self, output):
        if output not in self.outputs.names:
            raise KeyError("'{}' not in scenario outputs".format(output))

    def simulate(self, data):
        """No-op, as the data is assumed already available in the store
        """
        return data


class ScenarioModelBuilder(object):

    def __init__(self, name):
        self.scenario = ScenarioModel(name)
        self.logger = getLogger(__name__)

    def construct(self, scenario_config):
        """Build a ScenarioModel

        Arguments
        ---------
        scenario_config: dict
        """
        self.scenario.scenario_set = scenario_config['scenario_set']
        self.scenario.scenario_name = scenario_config['name']
        facets = scenario_config['facets']

        for facet in facets:
            spatial = facet['spatial_resolution']
            temporal = facet['temporal_resolution']

            spatial_res = self.scenario.regions.get_entry(spatial)
            temporal_res = self.scenario.intervals.get_entry(temporal)

            name = facet['name']
            self.scenario.add_output(name,
                                     spatial_res,
                                     temporal_res,
                                     facet['units'])

    def finish(self):
        """Return the built ScenarioModel
        """
        return self.scenario



"""Parameter and ParameterList
"""
import collections.abc
from collections import OrderedDict


class Parameter():
    """Data class to represent a single parameter's metadata

    Parameters
    ----------
    name : str, optional
    description : str, optional
    absolute_range : tuple, optional
    suggested_range : tuple, optional
    default_value : float or int, optional
    units : str, optional

    Attributes
    ----------
    name : str
    description : str
    absolute_range : tuple
    suggested_range : tuple
    default_value : float or int
    units : str

    """
    def __init__(self, name='', description='', absolute_range=(0, 0),
                 suggested_range=(0, 0), default_value=0, units=''):
        self.name = name
        self.description = description
        self.absolute_range = absolute_range
        self.suggested_range = suggested_range
        self.default_value = default_value
        self.units = units

    @classmethod
    def from_dict(cls, data):
        """Create a Parameter from a dictionary of values

        Parameters
        ----------
        data : dict
            Dictionary with values for each attribute, like::

                {
                    'name': 'smart_meter_savings',
                    'description': 'The savings from smart meters',
                    'absolute_range': (0, 100),
                    'suggested_range': (3, 10),
                    'default_value': 3,
                    'units': '%'
                }

        """
        parameter = cls()
        parameter.name = data['name']
        parameter.description = data['description']
        parameter.absolute_range = data['absolute_range']
        parameter.suggested_range = data['suggested_range']
        parameter.default_value = data['default_value']
        parameter.units = data['units']
        return parameter

    def as_dict(self):
        """Return a dict of parameter data
        """
        return {
            'name': self.name,
            'description': self.description,
            'absolute_range': self.absolute_range,
            'suggested_range': self.suggested_range,
            'default_value': self.default_value,
            'units': self.units
        }

    def __eq__(self, other):
        return self.name == other.name and \
            self.description == other.description and \
            self.absolute_range == other.absolute_range and \
            self.suggested_range == other.suggested_range and \
            self.default_value == other.default_value and \
            self.units == other.units


class ParameterList(collections.abc.Mapping):
    """Collection of parameters

    Parameters
    ----------
    parameters : list of `smif.parameter.Parameter`, optional

    """

    def __init__(self, parameters=None):
        self._data = OrderedDict()

        if parameters is not None:
            for parameter in parameters:
                self.add_parameter(parameter)

    def add_parameter(self, parameter):
        """Add a parameter

        Parameters
        ----------
        parameter : `smif.parameter.Parameter` or dict
        """
        if isinstance(parameter, dict):
            parameter = Parameter.from_dict(parameter)

        if parameter.name in self._data:
            raise ValueError("Parameter already defined")

        self._data[parameter.name] = parameter

    def as_list(self):
        """Return a list of dicts of parameter data
        """
        return [param.as_dict() for param in self._data.values()]

    def __getitem__(self, key):
        return self._data[key]

    def __iter__(self):
        for key in self._data:
            yield key

    def __len__(self):
        return len(self._data)

    @property
    def names(self):
        """Returns the names of all the contained parameters
        """
        return list(self._data.keys())

    @property
    def defaults(self):
        """Default parameter values
        """
        return {
            parameter.name: parameter.default_value
            for parameter in self._data.values()
        }

    def overridden(self, new_values):
        """Override parameter values, falling back to defaults

        Parameters
        ----------
        new_values : dict
            Dict with keys matching parameter names, values to override the
            defaults
        """
        return {
            parameter_name: new_values.get(parameter_name, default_value)
            for parameter_name, default_value in self.defaults.items()
        }



"""Implements a composite scenario/sector model/system-of-systems model

Begin by declaring the atomic units (scenarios and sector models) which make up
a the composite system-of-systems model and then add these to the composite.
Declare dependencies by using the ``add_dependency()`` method, passing in a
reference to the source model object, and a pointer to the model output
and sink parameter name for the destination model.

Run the model by calling the ``simulate()`` method, passing in a dictionary
containing data for any free hanging model inputs, not linked through a
dependency. A fully defined SosModel should have no hanging model inputs, and
can therefore be called using ``simulate()`` with no arguments.

Responsibility for passing required data to the contained models lies with the
calling class. This means data is only ever passed one layer down.
This simplifies the interface, and allows as little or as much hiding of data,
dependencies and model inputs as required.

Example
-------
A very simple example with just one scenario:

>>> elec_scenario = ScenarioModel('scenario')
>>> elec_scenario.add_output('demand', 'national', 'annual', 'GWh')
>>> sos_model = SosModel('simple')
>>> sos_model.add_model(elec_scenario)
>>> sos_model.simulate(2010)
{'scenario': {'demand': array([123])}}

A more comprehensive example with one scenario and one scenario model:

>>> elec_scenario = ScenarioModel('scenario')
>>> elec_scenario.add_output('demand', 'national', 'annual', 'GWh')
>>> class EnergyModel(SectorModel):
...   def extract_obj(self):
...     pass
...   def initialise(self):
...     pass
...   def simulate(self, timestep, data):
...     return {self.name: {'cost': data['input'] * 2}}
...
>>> energy_model = EnergyModel('model')
>>> energy_model.add_input('input', 'national', 'annual', 'GWh')
>>> energy_model.add_dependency(elec_scenario, 'demand', 'input', lambda x: x)
>>> sos_model = SosModel('sos')
>>> sos_model.add_model(elec_scenario)
>>> sos_model.add_model(energy_model)
>>> sos_model.simulate(2010)
{'model': {'cost': array([[246]])}, 'scenario': {'demand': array([[123]])}}

"""
from abc import ABCMeta, abstractmethod
from logging import getLogger

from smif.convert.area import get_register as get_region_register
from smif.convert.interval import get_register as get_interval_register
from smif.convert.unit import get_register as get_unit_register
from smif.metadata import MetadataSet
from smif.model.dependency import Dependency
from smif.parameters import ParameterList


class Model(metaclass=ABCMeta):
    """Abstract class represents the interface used to implement the composite
    `SosModel` and leaf classes `SectorModel` and `Scenario`.

    Arguments
    ---------
    name : str
    inputs : smif.metadata.MetaDataSet
    outputs : smif.metadata.MetaDataSet

    """
    regions = get_region_register()
    intervals = get_interval_register()
    units = get_unit_register()

    def __init__(self, name):
        self.name = name
        self.description = ''
        self._inputs = MetadataSet([])
        self._outputs = MetadataSet([])
        self.deps = {}

        self._parameters = ParameterList()

        self.timesteps = []

        self.logger = getLogger(__name__)

    @property
    def inputs(self):
        """All model inputs defined at this layer

        Returns
        -------
        smif.metadata.MetadataSet
        """
        return self._inputs

    @property
    def outputs(self):
        """All model outputs defined at this layer

        Returns
        -------
        smif.metadata.MetadataSet

        """
        return self._outputs

    @property
    def free_inputs(self):
        """Returns the free inputs not linked to a dependency at this layer

        Free inputs are passed up to higher layers for deferred linkages to
        dependencies.

        Returns
        -------
        smif.metadata.MetadataSet
        """
        all_input_names = set(self.inputs.names)
        dep_input_names = set(dep.sink.name for dep in self.deps.values())
        free_input_names = all_input_names - dep_input_names

        return MetadataSet(self.inputs[name] for name in free_input_names)

    @abstractmethod
    def simulate(self, data):
        """Override to implement the generation of model results

        Generate ``results`` for ``timestep`` using ``data``

        Arguments
        ---------
        data: smif.data_layer.DataHandle
            Access state, parameter values, dependency inputs.
        """
        pass

    def add_dependency(self, source_model, source_name, sink_name, function=None):
        """Adds a dependency to the current `Model` object

        Arguments
        ---------
        source_model : `smif.composite.Model`
            A reference to the source `~smif.composite.Model` object
        source_name : string
            The name of the model_output defined in the `source_model`
        sink_name : string
            The name of a model_input defined in this object

        """
        if source_name not in source_model.outputs.names:
            msg = "Output '{}' is not defined in '{}' model"
            raise ValueError(msg.format(source_name, source_model.name))

        if sink_name in self.free_inputs.names:
            source = source_model.outputs[source_name]
            sink = self.inputs[sink_name]
            self.deps[sink_name] = Dependency(
                source_model,
                source,
                sink,
                function
            )
            msg = "Added dependency from '%s:%s' to '%s:%s'"
            self.logger.debug(msg, source_model.name, source_name, self.name, sink_name)

        else:
            if sink_name in self.inputs.names:
                raise NotImplementedError("Multiple source dependencies"
                                          " not yet implemented")

            msg = "Inputs: '%s'. Free inputs: '%s'."
            self.logger.debug(msg, self.inputs.names, self.free_inputs.names)
            msg = "Input '{}' is not defined in '{}' model"
            raise ValueError(msg.format(sink_name, self.name))

    def add_parameter(self, parameter_dict):
        """Add a parameter to the model

        Arguments
        ---------
        parameter_dict : dict
            Contains the keys ``name``, ``description``,  ``absolute_range``,
            ``suggested_range``, ``default_value``, ``units``
        """
        self._parameters.add_parameter(parameter_dict)

    @property
    def parameters(self):
        """A list of parameters

        Returns
        -------
        smif.parameters.ParameterList
        """
        return self._parameters


class CompositeModel(Model, metaclass=ABCMeta):
    """Override to implement models which contain models.

    Inherited by `smif.model.sos_model.SosModel` and
    `smif.model.model_set.ModelSet`
    """

    def __init__(self, name):
        super().__init__(name)
        self.models = {}

    @property
    def free_inputs(self):
        """Returns the free inputs not linked to a dependency at this layer

        For this composite :class:`~smif.model.CompositeModel` this includes
        the free_inputs from all contained smif.model.Model objects

        Free inputs are passed up to higher layers for deferred linkages to
        dependencies.

        Returns
        -------
        smif.metadata.MetadataSet
        """
        # free inputs of current layer
        free_inputs = super().free_inputs.metadata

        # free inputs of all contained models
        for model in self.models.values():
            free_inputs.extend(model.free_inputs.metadata)

        # compose a new MetadataSet containing the free inputs
        metadataset = MetadataSet(free_inputs)

        return metadataset

    @property
    def outputs(self):
        outputs = super().outputs.metadata
        for model in self.models.values():
            outputs.extend(model.outputs.metadata)

        return MetadataSet(outputs)


def element_before(element, list_):
    """Return the element before a given element in a list, or None if the
    given element is first or not in the list.
    """
    if element not in list_ or element == list_[0]:
        return None
    else:
        index = list_.index(element)
        return list_[index - 1]


def element_after(element, list_):
    """Return the element after a given element in a list, or None if the
    given element is last or not in the list.
    """
    if element not in list_ or element == list_[-1]:
        return None
    else:
        index = list_.index(element)
        return list_[index + 1]



import logging

from smif.convert.area import RegionSet
from smif.convert.area import get_register as get_region_register
from smif.convert.interval import IntervalSet
from smif.convert.interval import get_register as get_interval_register
from smif.convert.register import Register
from smif.convert.unit import get_register as get_unit_register
from smif.data_layer import DatafileInterface

LOGGER = logging.getLogger(__name__)

REGIONS = get_region_register()
INTERVALS = get_interval_register()
UNITS = get_unit_register()


def load_region_sets(handler):
    """Loads the region sets into the project registries

    Parameters
    ----------
    handler: :class:`smif.data_layer.DataInterface`

    """
    region_definitions = handler.read_region_definitions()
    for region_def in region_definitions:
        region_def_name = region_def['name']
        LOGGER.info("Reading in region definition %s", region_def_name)
        region_data = handler.read_region_definition_data(region_def_name)
        region_set = RegionSet(region_def_name, region_data)
        REGIONS.register(region_set)


def load_interval_sets(handler):
    """Loads the time-interval sets into the project registries

    Parameters
    ----------
    handler: :class:`smif.data_layer.DataInterface`

    """
    interval_definitions = handler.read_interval_definitions()
    for interval_def in interval_definitions:
        interval_name = interval_def['name']
        LOGGER.info("Reading in interval definition %s", interval_name)
        interval_data = handler.read_interval_definition_data(interval_name)
        interval_set = IntervalSet(interval_name, interval_data)
        INTERVALS.register(interval_set)


def load_units(handler):
    """Loads the units into the project registries

    Parameters
    ----------
    handler: :class:`smif.data_layer.DataInterface`
    """
    unit_file = handler.read_units_file_name()
    if unit_file is not None:
        LOGGER.info("Loading units in from %s", unit_file)
        UNITS.register(unit_file)


def load_resolution_sets(directory):
    """Loads the region, interval units resolution sets

    Arguments
    ---------
    directory: str
        Path to the project directory
    """
    handler = DatafileInterface(directory)
    Register.data_interface = handler
    load_region_sets(handler)
    load_interval_sets(handler)
    load_units(handler)



"""Register, NDimensionalRegister and ResolutionSet
abstract classes to contain area, interval and unit metadata.

Implemented by :class:`smif.convert.interval.TimeIntervalRegister`,
:class:`smif.convert.area.RegionRegister` and
:class:`smif.convert.unit.UnitRegister`.
"""
import logging
from abc import ABCMeta, abstractmethod
from collections import OrderedDict, defaultdict

import numpy as np


class LogMixin(object):

    @property
    def logger(self):
        name = '.'.join([__name__, self.__class__.__name__])
        return logging.getLogger(name)


class Register(LogMixin, metaclass=ABCMeta):
    """Abstract class which holds the ResolutionSets

    Arguments
    ---------
    axis : int, default=None
        The axis over which operations on the data array are performed

    """
    data_interface = None

    def __init__(self, axis=None):
        self.axis = axis

    @property
    @abstractmethod
    def names(self):
        raise NotImplementedError

    @abstractmethod
    def register(self, resolution_set):
        raise NotImplementedError

    @abstractmethod
    def get_coefficients(self, source, destination):
        raise NotImplementedError

    def convert(self, data, from_set_name, to_set_name):
        """Convert a list of data points for a given set to another set

        Parameters
        ----------
        data: numpy.ndarray
        from_set_name: str
        to_set_name: str

        Returns
        -------
        numpy.ndarray

        """
        coefficients = self.get_coefficients(from_set_name, to_set_name)

        if self.axis is not None:
            data_count = data.shape[self.axis]
            if coefficients.shape[0] != data_count:
                msg = "Size of coefficient array does not match source " \
                      "resolution set from data matrix: %s != %s"
                raise ValueError(msg, coefficients.shape[self.axis],
                                 data_count)

        if self.axis == 0:
            converted = np.dot(coefficients.T, data)
        elif self.axis == 1:
            converted = np.dot(data, coefficients)
        else:
            converted = np.dot(data, coefficients)

        self.logger.debug("Converting from %s to %s.", from_set_name, to_set_name)
        self.logger.debug("Converted value from %s to %s", data.sum(), converted.sum() )

        return converted


class NDimensionalRegister(Register):
    """Abstract class which holds N-Dimensional ResolutionSets

    Arguments
    ---------
    axis : int, default=None
        The axis over which operations on the data array are performed

    """
    def __init__(self, axis=None):
        super().__init__(axis)
        self._register = OrderedDict()
        self._conversions = defaultdict(dict)

    def register(self, resolution_set):
        """Add a ResolutionSet to the register

        Parameters
        ----------
        resolution_set : :class:`smif.convert.ResolutionSet`

        Raises
        ------
        ValueError
            If a ResolutionSet of the same name already exists in the register
        """
        if resolution_set.name in self._register:
            msg = "A ResolutionSet named {} has already been loaded"
            raise ValueError(msg.format(resolution_set.name))

        self.logger.info("Registering '%s' with %i items",
                         resolution_set.name,
                         len(resolution_set))

        self._register[resolution_set.name] = resolution_set

    @property
    def names(self):
        """Names of registered region sets

        Returns
        -------
        sets: list of str
        """
        return list(self._register.keys())

    def get_entry(self, name):
        """Returns the ResolutionSet of `name`

        Arguments
        ---------
        name : str
            The unique identifier of a ResolutionSet in the register

        Returns
        -------
        smif.convert.ResolutionSet

        """
        if name not in self._register:
            msg = "ResolutionSet '{}' not registered"
            raise ValueError(msg.format(name))
        return self._register[name]

    def _write_coefficients(self, source, destination, data):
        if self.data_interface:
            self.data_interface.write_coefficients(source,
                                                   destination,
                                                   data)
        else:
            msg = "Data interface not available to write coefficients"
            self.logger.warning(msg)

    def get_coefficients(self, source, destination):
        """Get coefficients representing intersection of sets

        Arguments
        ---------
        source : string
            The name of the source set
        destination : string
            The name of the destination set

        Returns
        -------
        numpy.ndarray
        """
        from_set = self.get_entry(source)
        to_set = self.get_entry(destination)

        if from_set.coverage != to_set.coverage:
            log_msg = "Coverage for '%s' is %d and does not match coverage " \
                    "for '%s' which is %d"
            self.logger.warning(log_msg, from_set.name, from_set.coverage,
                                to_set.name, to_set.coverage)

        if self.data_interface:

            self.logger.info("Using data interface to load coefficients")
            coefficients = self.data_interface.read_coefficients(source,
                                                                 destination)

            if coefficients is None:
                msg = "Coefficients not found, generating coefficients for %s to %s"
                self.logger.info(msg, source, destination)

                coefficients = self._obtain_coefficients(from_set, to_set)
                self._write_coefficients(source, destination, coefficients)

        else:

            msg = "No data interface specified, generating coefficients for %s to %s"
            self.logger.info(msg, source, destination)
            coefficients = self._obtain_coefficients(from_set, to_set)

        return coefficients

    def _obtain_coefficients(self, from_set, to_set):
        """
        """
        coefficients = np.zeros((len(from_set), len(to_set)), dtype=np.float)
        self.logger.debug("Coefficients array is of shape %s for %s to %s",
                          coefficients.shape, from_set.name, to_set.name)

        from_names = from_set.get_entry_names()
        for to_idx, to_entry in enumerate(to_set):
            for from_idx in from_set.intersection(to_entry):

                proportion = from_set.get_proportion(from_idx, to_entry)

                self.logger.debug("%i percent of %s is in %s",
                                  proportion * 100,
                                  to_entry.name, from_set.data[from_idx].name)
                from_idx = from_names.index(from_set.data[from_idx].name)

                coefficients[from_idx, to_idx] = proportion
        return coefficients


class ResolutionSet(metaclass=ABCMeta):
    """Abstract class which holds the Resolution definitions
    """
    def __init__(self):

        self.name = ''
        self.description = ''
        self.data = []
        self.logger = logging.getLogger(__name__)

    def as_dict(self):
        """Get a serialisable representation of the object
        """
        return {'name': self.name,
                'description': self.description}

    def __iter__(self):
        return iter(self.data)

    @abstractmethod
    def get_entry_names(self):
        """Get the names of the entries in the ResolutionSet

        Returns
        -------
        set
            The set of names which identify each entry in the ResolutionSet
        """
        raise NotImplementedError

    @abstractmethod
    def intersection(self, bounds):
        """Return the subset of entries intersecting with the bounds
        """
        raise NotImplementedError

    @abstractmethod
    def get_proportion(self, entry_a, entry_b):
        """Calculate the proportion of `entry_a` and `entry_b`

        Arguments
        ---------
        entry_a : string
            Name of an entry in `ResolutionSet`
        entry_b : string
            Name of an entry in `ResolutionSet`

        Returns
        -------
        float
            The proportion of `entry_a` and `entry_b`
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def coverage(self):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_bounds(entry):
        """Implement this helper method to return bounds from an entry in the register

        Arguments
        ---------
        entry
            An entry from a ResolutionSet

        Returns
        -------
        bounds
            The bounds of the entry
        """
        raise NotImplementedError



"""File-backed data interface
"""
import csv
import glob
import os
import re
from csv import DictReader

import fiona
import pyarrow as pa
from smif.data_layer.data_interface import (DataExistsError, DataInterface,
                                            DataMismatchError,
                                            DataNotFoundError)
from smif.data_layer.load import dump, load


class DatafileInterface(DataInterface):
    """Read and write interface to YAML / CSV configuration files
    and intermediate CSV / native-binary data storage.

    Project.yml

    Arguments
    ---------
    base_folder: str
        The path to the configuration and data files
    storage_format: str
        The format used to store intermediate data (local_csv, local_binary)
    """
    def __init__(self, base_folder, storage_format='local_binary'):
        super().__init__()

        self.base_folder = base_folder
        self.storage_format = storage_format

        self.file_dir = {}
        self.file_dir['project'] = os.path.join(base_folder, 'config')
        self.file_dir['results'] = os.path.join(base_folder, 'results')

        config_folders = {
            'sos_model_runs': 'config',
            'sos_models': 'config',
            'sector_models': 'config',
            'initial_conditions': 'data',
            'interval_definitions': 'data',
            'interventions': 'data',
            'narratives': 'data',
            'region_definitions': 'data',
            'scenarios': 'data',
            'coefficients': 'data',
            'strategies': 'data'
        }

        for category, folder in config_folders.items():
            dirname = os.path.join(base_folder, folder, category)
            # ensure each directory exists
            os.makedirs(dirname, exist_ok=True)
            # store dirname
            self.file_dir[category] = dirname

    def read_units_file_name(self):
        project_config = self._read_project_config()
        filename = project_config['units']
        self.logger.debug("Units filename is %s", filename)
        if filename is not None:
            path = os.path.join(self.base_folder, 'data')
            units_file_path = os.path.join(path, filename)
            if os.path.isfile(units_file_path):
                return units_file_path
            else:
                return None
        else:
            return None

    def read_sos_model_runs(self):
        """Read all system-of-system model runs from Yaml files

        sos_model_runs.yml

        Returns
        -------
        list
            A list of sos_model_run dicts
        """
        sos_model_runs = []

        sos_model_run_names = self._read_filenames_in_dir(self.file_dir['sos_model_runs'],
                                                          '.yml')
        for sos_model_run_name in sos_model_run_names:
            sos_model_runs.append(self._read_yaml_file(
                self.file_dir['sos_model_runs'], sos_model_run_name))

        return sos_model_runs

    def read_sos_model_run(self, sos_model_run_name):
        """Read a system-of-system model run

        Arguments
        ---------
        sos_model_run_name: str
            A sos_model_run name

        Returns
        -------
        sos_model_run: dict
            A sos_model_run dictionary
        """
        if not self._sos_model_run_exists(sos_model_run_name):
            raise DataNotFoundError("sos_model_run '%s' not found" % sos_model_run_name)
        return self._read_yaml_file(self.file_dir['sos_model_runs'], sos_model_run_name)

    def _sos_model_run_exists(self, name):
        return os.path.exists(
            os.path.join(self.file_dir['sos_model_runs'], name + '.yml'))

    def write_sos_model_run(self, sos_model_run):
        """Write system-of-system model run to Yaml file

        Arguments
        ---------
        sos_model_run: dict
            A sos_model_run dictionary
        """
        if self._sos_model_run_exists(sos_model_run['name']):
            raise DataExistsError("sos_model_run '%s' already exists" % sos_model_run['name'])
        else:
            self._write_yaml_file(self.file_dir['sos_model_runs'],
                                  sos_model_run['name'], sos_model_run)

    def update_sos_model_run(self, sos_model_run_name, sos_model_run):
        """Update system-of-system model run in Yaml file

        Arguments
        ---------
        sos_model_run_name: str
            A sos_model_run name
        sos_model_run: dict
            A sos_model_run dictionary
        """
        if sos_model_run_name != sos_model_run['name']:
            raise DataMismatchError(
                "sos_model_run name '{}' must match '{}'".format(
                    sos_model_run_name,
                    sos_model_run['name']))

        if not self._sos_model_run_exists(sos_model_run_name):
            raise DataNotFoundError("sos_model_run '%s' not found" % sos_model_run_name)
        self._write_yaml_file(self.file_dir['sos_model_runs'],
                              sos_model_run['name'], sos_model_run)

    def delete_sos_model_run(self, sos_model_run_name):
        """Delete a system-of-system model run

        Arguments
        ---------
        sos_model_run_name: str
            A sos_model_run name
        """
        if not self._sos_model_run_exists(sos_model_run_name):
            raise DataNotFoundError("sos_model_run '%s' not found" % sos_model_run_name)

        os.remove(os.path.join(self.file_dir['sos_model_runs'], sos_model_run_name + '.yml'))

    def read_sos_models(self):
        """Read all system-of-system models from Yaml files

        Returns
        -------
        list
            A list of sos_models dicts
        """
        sos_models = []

        sos_model_names = self._read_filenames_in_dir(self.file_dir['sos_models'], '.yml')
        for sos_model_name in sos_model_names:
            sos_models.append(self._read_yaml_file(self.file_dir['sos_models'],
                                                   sos_model_name))
        return sos_models

    def read_sos_model(self, sos_model_name):
        """Read a specific system-of-system model

        Arguments
        ---------
        sos_model_name: str
            A sos_model name

        Returns
        -------
        sos_model: dict
            A sos_model dictionary
        """
        if not self._sos_model_exists(sos_model_name):
            raise DataNotFoundError("sos_model '%s' not found" % sos_model_name)
        return self._read_yaml_file(self.file_dir['sos_models'], sos_model_name)

    def _sos_model_exists(self, name):
        return os.path.exists(
            os.path.join(self.file_dir['sos_models'], name + '.yml'))

    def write_sos_model(self, sos_model):
        """Write system-of-system model to Yaml file

        Arguments
        ---------
        sos_model: dict
            A sos_model dictionary
        """
        if self._sos_model_exists(sos_model['name']):
            raise DataExistsError("sos_model '%s' already exists" % sos_model['name'])
        else:
            self._write_yaml_file(
                self.file_dir['sos_models'],
                sos_model['name'],
                sos_model
            )

    def update_sos_model(self, sos_model_name, sos_model):
        """Update system-of-system model in Yaml file

        Arguments
        ---------
        sos_model_name: str
            A sos_model name
        sos_model: dict
            A sos_model dictionary
        """
        if sos_model_name != sos_model['name']:
            raise DataMismatchError(
                "sos_model name '{}' must match '{}'".format(
                    sos_model_name,
                    sos_model['name']))

        if not self._sos_model_exists(sos_model_name):
            raise DataNotFoundError("sos_model '%s' not found" % sos_model_name)
        self._write_yaml_file(self.file_dir['sos_models'],
                              sos_model['name'], sos_model)

    def delete_sos_model(self, sos_model_name):
        """Delete a system-of-system model run

        Arguments
        ---------
        sos_model_name: str
            A sos_model name
        """
        if not self._sos_model_exists(sos_model_name):
            raise DataNotFoundError("sos_model '%s' not found" % sos_model_name)

        os.remove(os.path.join(self.file_dir['sos_models'], sos_model_name + '.yml'))

    def read_sector_models(self):
        """Read all sector models from Yaml files

        sector_models.yml

        Returns
        -------
        list
            A list of sector_model dicts
        """
        sector_models = []
        file_dir = self.file_dir['sector_models']

        sector_model_names = self._read_filenames_in_dir(file_dir, '.yml')
        for sector_model_name in sector_model_names:
            sector_models.append(self._read_yaml_file(file_dir, sector_model_name))

        return sector_models

    def _sector_model_exists(self, name):
        file_name = name + '.yml'
        file_dir = self.file_dir['sector_models']

        if os.path.exists(os.path.join(file_dir, file_name)):
            return self._read_yaml_file(file_dir, name)

    def read_sector_model(self, sector_model_name):
        """Read a sector model

        Arguments
        ---------
        sector_model_name: str
            A sector_model name

        Returns
        -------
        sector_model: dict
            A sector_model dictionary
        """
        sector_model = self._sector_model_exists(sector_model_name)
        if sector_model is None:
            raise DataNotFoundError("sector_model '%s' not found" % sector_model_name)

        return sector_model

    def write_sector_model(self, sector_model):
        """Write sector model to Yaml file

        Arguments
        ---------
        sector_model: dict
            A sector_model dictionary
        """
        if self._sector_model_exists(sector_model['name']):
            raise DataExistsError("sector_model '%s' already exists" % sector_model['name'])

        self._write_yaml_file(
            self.file_dir['sector_models'], sector_model['name'], sector_model)

    def update_sector_model(self, sector_model_name, sector_model):
        """Update sector model in Yaml file

        Arguments
        ---------
        sector_model_name: str
            A sector_model name
        sector_model: dict
            A sector_model dictionary
        """
        if sector_model_name != sector_model['name']:
            raise DataMismatchError(
                "sector_model name '{}' must match '{}'".format(
                    sector_model_name,
                    sector_model['name']))

        if not self._sector_model_exists(sector_model_name):
            raise DataNotFoundError("sector_model '%s' not found" % sector_model_name)

        self._write_yaml_file(self.file_dir['sector_models'],
                              sector_model['name'], sector_model)

    def delete_sector_model(self, sector_model_name):
        """Delete a sector model

        Arguments
        ---------
        sector_model_name: str
            A sector_model name
        """
        if not self._sector_model_exists(sector_model_name):
            raise DataNotFoundError("sector_model '%s' not found" % sector_model_name)

        os.remove(os.path.join(self.file_dir['sector_models'], sector_model_name + '.yml'))

    def read_interventions(self, filename):
        """Read the interventions from filename

        Arguments
        ---------
        filename: str
            The name of the intervention yml file to read in
        """
        filepath = self.file_dir['interventions']
        return self._read_yaml_file(filepath, filename, extension='')

    def read_strategies(self, filename):
        """Read the strategy data from filename

        Arguments
        ---------
        filename: str
            The name of the strategy yml file to read in
        """
        filepath = self.file_dir['strategies']
        return self._read_yaml_file(filepath, filename, extension='')

    def read_initial_conditions(self, filename):
        """Read the initial conditions from filename

        Arguments
        ---------
        filename: str
            The name of the initial conditions yml file to read in
        """
        filepath = self.file_dir['initial_conditions']
        return self._read_yaml_file(filepath, filename, extension='')

    def read_region_definitions(self):
        """Read region_definitions from project configuration

        Returns
        -------
        list
            A list of region_definition dicts
        """
        project_config = self._read_project_config()
        return project_config['region_definitions']

    def _region_definition_exists(self, region_definition_name):
        for region_definition in self.read_region_definitions():
            if region_definition['name'] == region_definition_name:
                return region_definition

    def read_region_definition_data(self, region_definition_name):
        """Read region_definition data file into a Fiona feature collection

        The file format must be possible to parse with GDAL, and must contain
        an attribute "name" to use as an identifier for the region_definition.

        Arguments
        ---------
        region_definition_name: str
            Name of the region_definition

        Returns
        -------
        list
            A list of data from the specified file in a fiona formatted dict
        """
        # Find filename for this region_definition_name
        region_definition = self._region_definition_exists(region_definition_name)

        if region_definition is None:
            raise DataNotFoundError(
                "Region definition '{}' not found".format(region_definition_name))
        else:
            filename = region_definition['filename']

        # Read the region data from file
        filepath = os.path.join(self.file_dir['region_definitions'], filename)
        with fiona.drivers():
            with fiona.open(filepath) as src:
                data = [f for f in src]

        return data

    def read_region_names(self, region_definition_name):
        """Return the set of unique region names in region set `region_definition_name`
        """
        names = []
        for feature in self.read_region_definition_data(region_definition_name):
            if isinstance(feature['properties']['name'], str):
                if feature['properties']['name'].isdigit():
                    names.append(int(feature['properties']['name']))
                else:
                    names.append(feature['properties']['name'])
            else:
                names.append(feature['properties']['name'])
    
        return names

    def write_region_definition(self, region_definition):
        """Write region_definition to project configuration

        Arguments
        ---------
        region_definition: dict
            A region_definition dict
        """

        if self._region_definition_exists(region_definition['name']):
            raise DataExistsError(
                "region_definition '%s' already exists" % region_definition['name'])

        project_config = self._read_project_config()
        project_config['region_definitions'].append(region_definition)
        self._write_project_config(project_config)

    def update_region_definition(self, region_definition_name, region_definition):
        """Update region_definition to project configuration

        Arguments
        ---------
        region_definition_name: str
            Name of the (original) entry
        region_definition: dict
            The updated region_definition dict
        """
        previous_region_definition = self._region_definition_exists(region_definition_name)
        if previous_region_definition is None:
            raise DataNotFoundError(
                "region_definition '%s' does not exist" % region_definition_name)

        # Update
        project_config = self._read_project_config()

        idx = None
        for i, existing_region_definition in enumerate(project_config['region_definitions']):
            if existing_region_definition['name'] == region_definition_name:
                # Guaranteed to match thanks to existence check above
                idx = i
                break

        project_config['region_definitions'][idx] = region_definition

        self._write_project_config(project_config)

    def read_interval_definitions(self):
        """Read interval_definition sets from project configuration

        Returns
        -------
        list
            A list of interval_definition set dicts
        """
        project_config = self._read_project_config()
        return project_config['interval_definitions']

    def _interval_definition_exists(self, interval_definition_name):
        for interval_definition in self.read_interval_definitions():
            if interval_definition['name'] == interval_definition_name:
                return interval_definition

    def read_interval_definition_data(self, interval_definition_name):
        """Read data for an interval definition

        Arguments
        ---------
        interval_definition_name: str

        Returns
        -------
        dict
            Interval definition data

        Notes
        -----
        Expects csv file to contain headings of `id`, `start`, `end`
        """
        interval_definition = self._interval_definition_exists(interval_definition_name)

        if interval_definition is None:
            raise DataNotFoundError("Interval set definition '{}' does not exist".format(
                interval_definition_name))

        filename = interval_definition['filename']
        filepath = os.path.join(self.file_dir['interval_definitions'], filename)

        names = {}

        with open(filepath, 'r') as csvfile:
            reader = DictReader(csvfile)
            data = []
            for interval in reader:

                if interval['id'].isdigit():
                    name = int(interval['id'])
                else:
                    name = interval['id']
                interval_tuple = (interval['start'], interval['end'])
                if name in names:
                    # Append duration to existing entry
                    data[names[name]][1].append(interval_tuple)
                else:
                    # Make a new entry
                    data.append((name, [interval_tuple]))
                    names[name] = len(data) - 1

        return data

    def read_interval_names(self, interval_definition_name):
        return [
            interval[0]
            for interval
            in self.read_interval_definition_data(interval_definition_name)
        ]

    def write_interval_definition(self, interval_definition):
        """Write interval_definition to project configuration

        Arguments
        ---------
        interval_definition: dict
            A interval_definition dict
        """
        if self._interval_definition_exists(interval_definition['name']):
            raise DataExistsError(
                "Interval definition '%s' already exists" % interval_definition['name'])

        project_config = self._read_project_config()
        project_config['interval_definitions'].append(interval_definition)
        self._write_project_config(project_config)

    def update_interval_definition(self, interval_definition_name, interval_definition):
        """Update interval_definition to project configuration

        Arguments
        ---------
        interval_definition_name: str
            Name of the (original) entry
        interval_definition: dict
            The updated interval_definition dict
        """
        if not self._interval_definition_exists(interval_definition_name):
            raise DataNotFoundError(
                "Interval definition '%s' does not exist" % interval_definition_name)

        project_config = self._read_project_config()

        # Create updated list
        idx = None
        for i, existing_interval_def in enumerate(project_config['interval_definitions']):
            if existing_interval_def['name'] == interval_definition_name:
                # Guaranteed to match thanks to existence check above
                idx = i
                break

        project_config['interval_definitions'][idx] = interval_definition

        self._write_project_config(project_config)

    def read_scenario_definitions(self):
        project_config = self._read_project_config()
        return project_config['scenarios']

    def _scenario_definition_exists(self, scenario_name):
        for scenario in self.read_scenario_definitions():
            if scenario['name'] == scenario_name:
                return scenario

    def read_scenario_set_scenario_definitions(self, scenario_set_name):
        """Read all scenarios from a certain scenario_set

        Arguments
        ---------
        scenario_set_name: str
            Name of the scenario_set

        Returns
        -------
        list
            A list of scenarios within the specified 'scenario_set_name'
        """
        # Filter only the scenarios of the selected scenario_set_name
        filtered_scenario_data = [
            data for data in self.read_scenario_definitions()
            if data['scenario_set'] == scenario_set_name
        ]

        if not filtered_scenario_data:
            self.logger.warning(
                "Scenario set '%s' has no scenarios defined", scenario_set_name)

        return filtered_scenario_data

    def read_scenario_definition(self, scenario_name):
        """Read scenario definition data

        Arguments
        ---------
        scenario_name: str
            Name of the scenario

        Returns
        -------
        dict
            The scenario definition
        """
        scenario_definition = self._scenario_definition_exists(scenario_name)

        if scenario_definition is None:
            raise DataNotFoundError(
                "Scenario definition '{}' not found".format(scenario_name))

        return scenario_definition

    def read_scenario_sets(self):
        """Read scenario sets from project configuration

        Returns
        -------
        list
            A list of scenario set dicts
        """
        project_config = self._read_project_config()
        return project_config['scenario_sets']

    def _scenario_set_exists(self, scenario_set_name):
        for scenario_set in self.read_scenario_sets():
            if scenario_set['name'] == scenario_set_name:
                return scenario_set

    def read_scenario_set(self, scenario_set_name):
        """Read a scenario_set

        Arguments
        ---------
        scenario_set_name: str
            Name of the scenario_set

        Returns
        -------
        dict
            Scenario set definition
        """
        scenario_set = self._scenario_set_exists(scenario_set_name)

        if scenario_set is None:
            raise DataNotFoundError(
                "Scenario set '{}' not found".format(scenario_set_name))

        return scenario_set

    def write_scenario_set(self, scenario_set):
        """Write scenario_set to project configuration

        Arguments
        ---------
        scenario_set: dict
            A scenario_set dict
        """
        if self._scenario_set_exists(scenario_set['name']):
            raise DataExistsError("scenario_set '%s' already exists" % scenario_set['name'])

        project_config = self._read_project_config()
        project_config['scenario_sets'].append(scenario_set)
        self._write_project_config(project_config)

    def update_scenario_set(self, scenario_set_name, scenario_set):
        """Update scenario_set to project configuration

        Arguments
        ---------
        scenario_set_name: str
            Name of the (original) entry
        scenario_set: dict
            The updated scenario_set dict
        """
        if not self._scenario_set_exists(scenario_set_name):
            raise DataNotFoundError("scenario_set '%s' not found" % scenario_set_name)

        project_config = self._read_project_config()

        idx = None
        for i, existing_scenario_set in enumerate(project_config['scenario_sets']):
            if existing_scenario_set['name'] == scenario_set_name:
                # Guaranteed to find a match thanks to existence assertion above
                idx = i
                break

        project_config['scenario_sets'][idx] = scenario_set

        self._write_project_config(project_config)

    def delete_scenario_set(self, scenario_set_name):
        """Delete scenario_set from project configuration
        and all scenarios within scenario_set

        Arguments
        ---------
        scenario_set_name: str
            A scenario_set name
        """
        if not self._scenario_set_exists(scenario_set_name):
            raise DataNotFoundError("scenario_set '%s' not found" % scenario_set_name)

        project_config = self._read_project_config()

        project_config['scenarios'] = [
            entry for entry in project_config['scenarios']
            if entry['scenario_set'] != scenario_set_name
        ]

        project_config['scenario_sets'] = [
            entry for entry in project_config['scenario_sets']
            if (entry['name'] != scenario_set_name)
        ]

        self._write_project_config(project_config)

    def read_scenarios(self):
        """Read scenarios from project configuration

        Returns
        -------
        list
            A list of scenario dicts
        """
        project_config = self._read_project_config()
        return project_config['scenarios']

    def read_scenario(self, scenario_name):
        """Read a scenario

        Arguments
        ---------
        scenario_name: str
            Name of the scenario

        Returns
        -------
        dict
            A scenario dictionary
        """
        project_config = self._read_project_config()
        for scenario_data in project_config['scenarios']:
            if scenario_data['name'] == scenario_name:
                return scenario_data
        raise DataNotFoundError("scenario '%s' not found" % scenario_name)

    def _scenario_exists(self, scenario_name):
        project_config = self._read_project_config()
        for scenario in project_config['scenarios']:
            if scenario['name'] == scenario_name:
                return scenario

    def write_scenario(self, scenario):
        """Write scenario to project configuration

        Arguments
        ---------
        scenario: dict
            A scenario dict
        """
        if self._scenario_exists(scenario['name']):
            raise DataExistsError("scenario '%s' already exists" % scenario['name'])
        else:
            project_config = self._read_project_config()
            project_config['scenarios'].append(scenario)
            self._write_project_config(project_config)

    def update_scenario(self, scenario_name, scenario):
        """Update scenario to project configuration

        Arguments
        ---------
        scenario_name: str
            Name of the (original) entry
        scenario: dict
            The updated scenario dict
        """
        if not self._scenario_exists(scenario_name):
            raise DataNotFoundError("scenario '%s' not found" % scenario_name)

        project_config = self._read_project_config()

        idx = None
        for i, existing_scenario in enumerate(project_config['scenarios']):
            if existing_scenario['name'] == scenario_name:
                # Guaranteed to match given existence check above
                idx = i
                break

        project_config['scenarios'][idx] = scenario

        self._write_project_config(project_config)

    def delete_scenario(self, scenario_name):
        """Delete scenario from project configuration

        Arguments
        ---------
        scenario_name: str
            A scenario name
        """
        if not self._scenario_exists(scenario_name):
            raise DataNotFoundError("scenario '%s' not found" % scenario_name)

        project_config = self._read_project_config()

        project_config['scenarios'] = [
            entry for entry in project_config['scenarios']
            if (entry['name'] != scenario_name)
        ]

        self._write_project_config(project_config)

    def read_scenario_data(self, scenario_name, facet_name,
                           spatial_resolution, temporal_resolution, timestep):
        """Read scenario data file

        Arguments
        ---------
        scenario_name: str
            Name of the scenario
        facet_name: str
            Name of the scenario facet to read
        spatial_resolution : str
        temporal_resolution : str
        timestep: int

        Returns
        -------
        data: numpy.ndarray

        """
        # Find filenames for this scenario
        filename = None
        project_config = self._read_project_config()
        for scenario_data in project_config['scenarios']:
            if scenario_data['name'] == scenario_name:
                for facet in scenario_data['facets']:
                    if facet['name'] == facet_name:
                        filename = facet['filename']
                        break
                break

        if filename is None:
            raise DataNotFoundError(
                "Scenario '{}' with facet '{}' not found".format(
                    scenario_name, facet_name))

        # Read the scenario data from file
        filepath = os.path.join(self.file_dir['scenarios'], filename)
        data = [
            datum for datum in
            self._get_data_from_csv(filepath)
            if int(datum['year']) == timestep
        ]

        # Position of names in these lists dictates position of
        # data in data array
        region_names = self.read_region_names(spatial_resolution)
        interval_names = self.read_interval_names(temporal_resolution)

        return self.data_list_to_ndarray(data, region_names, interval_names)

    def read_narrative_sets(self):
        """Read narrative sets from project configuration

        Returns
        -------
        list
            A list of narrative set dicts
        """
        # Find filename for this narrative
        project_config = self._read_project_config()
        return project_config['narrative_sets']

    def read_narrative_set(self, narrative_set_name):
        """Read all narratives from a certain narrative_set

        Arguments
        ---------
        narrative_set_name: str
            Name of the narrative_set

        Returns
        -------
        list
            A narrative_set dictionary
        """
        project_config = self._read_project_config()
        for narrative_data in project_config['narrative_sets']:
            if narrative_data['name'] == narrative_set_name:
                return narrative_data
        raise DataNotFoundError("narrative_set '%s' not found" % narrative_set_name)

    def _narrative_set_exists(self, narrative_set_name):
        project_config = self._read_project_config()
        for narrative_set in project_config['narrative_sets']:
            if narrative_set['name'] == narrative_set_name:
                return narrative_set

    def write_narrative_set(self, narrative_set):
        """Write narrative_set to project configuration

        Arguments
        ---------
        narrative_set: dict
            A narrative_set dict
        """
        if self._narrative_set_exists(narrative_set['name']):
            raise DataExistsError("narrative_set '%s' already exists" % narrative_set['name'])
        else:
            project_config = self._read_project_config()
            project_config['narrative_sets'].append(narrative_set)
            self._write_project_config(project_config)

    def update_narrative_set(self, narrative_set_name, narrative_set):
        """Update narrative_set to project configuration

        Arguments
        ---------
        narrative_set_name: str
            Name of the (original) entry
        narrative_set: dict
            The updated narrative_set dict
        """
        if not self._narrative_set_exists(narrative_set_name):
            raise DataNotFoundError("narrative_set '%s' not found" % narrative_set_name)

        project_config = self._read_project_config()

        idx = None
        for i, existing_narrative_set in enumerate(project_config['narrative_sets']):
            if existing_narrative_set['name'] == narrative_set_name:
                # Guaranteed to match thanks to existence check above
                idx = i
                break

        project_config['narrative_sets'][idx] = narrative_set

        self._write_project_config(project_config)

    def delete_narrative_set(self, narrative_set_name):
        """Delete narrative_set from project configuration

        Arguments
        ---------
        narrative_set_name: str
            A narrative_set name
        """
        if not self._narrative_set_exists(narrative_set_name):
            raise DataNotFoundError("narrative_set '%s' not found" % narrative_set_name)

        project_config = self._read_project_config()

        project_config['narrative_sets'] = [
            entry for entry in project_config['narrative_sets']
            if (entry['name'] != narrative_set_name)
        ]

        self._write_project_config(project_config)

    def read_narratives(self):
        """Read narrative sets from project configuration

        Returns
        -------
        list
            A list of narrative set dicts
        """
        project_config = self._read_project_config()
        return project_config['narratives']

    def read_narrative(self, narrative_name):
        """Read all narratives from a certain narrative

        Arguments
        ---------
        narrative_name: str
            Name of the narrative

        Returns
        -------
        list
            A narrative dictionary
        """
        project_config = self._read_project_config()
        for narrative_data in project_config['narratives']:
            if narrative_data['name'] == narrative_name:
                return narrative_data
        raise DataNotFoundError("narrative '%s' not found" % narrative_name)

    def _narrative_exists(self, narrative_name):
        project_config = self._read_project_config()
        for narrative in project_config['narratives']:
            if narrative['name'] == narrative_name:
                return narrative

    def write_narrative(self, narrative):
        """Write narrative to project configuration

        Arguments
        ---------
        narrative: dict
            A narrative dict
        """
        if self._narrative_exists(narrative['name']):
            raise DataExistsError("narrative '%s' already exists" % narrative['name'])
        else:
            project_config = self._read_project_config()
            project_config['narratives'].append(narrative)
            self._write_project_config(project_config)

    def update_narrative(self, narrative_name, narrative):
        """Update narrative to project configuration

        Arguments
        ---------
        narrative_name: str
            Name of the (original) entry
        narrative: dict
            The updated narrative dict
        """
        if not self._narrative_exists(narrative_name):
            raise DataNotFoundError("narrative '%s' not found" % narrative_name)

        project_config = self._read_project_config()

        idx = None
        for i, existing_narrative in enumerate(project_config['narratives']):
            if existing_narrative['name'] == narrative_name:
                # Guaranteed to match given existence check
                idx = i
                break

        project_config['narratives'][idx] = narrative

        self._write_project_config(project_config)

    def delete_narrative(self, narrative_name):
        """Delete narrative from project configuration

        Arguments
        ---------
        narrative_name: str
            A narrative name
        """
        if not self._narrative_exists(narrative_name):
            raise DataNotFoundError("narrative '%s' not found" % narrative_name)

        project_config = self._read_project_config()

        project_config['narratives'] = [
            entry for entry in project_config['narratives']
            if (entry['name'] != narrative_name)
        ]

        self._write_project_config(project_config)

    def read_narrative_data(self, narrative_name):
        """Read narrative data file

        Arguments
        ---------
        narrative_name: str
            Name of the narrative

        Returns
        -------
        list
            A list with dictionaries containing the contents of 'narrative_name' data file
        """
        # Find filename for this narrative
        filename = None
        project_config = self._read_project_config()
        for narrative in project_config['narratives']:
            if narrative['name'] == narrative_name:
                filename = narrative['filename']
                break

        if filename is None:
            raise DataNotFoundError(
                'Narrative \'{}\' has no data defined'.format(narrative_name))

        # Read the narrative data from file
        try:
            narrative_data = load(os.path.join(self.file_dir['narratives'], filename))
        except FileNotFoundError:
            raise DataNotFoundError(
                'Narrative \'{}\' has no data defined'.format(narrative_name))

        return narrative_data

    def read_narrative_definition(self, narrative_name):
        """Read the narrative definition

        Arguments
        ---------
        narrative_name: str
            Name of the narrative

        Returns
        -------
        dict

        """
        project_config = self._read_project_config()
        for narrative in project_config['narratives']:
            if narrative['name'] == narrative_name:
                return narrative

        raise DataNotFoundError('Narrative \'{}\' not found'.format(narrative_name))

    def read_coefficients(self, source_name, destination_name):
        """Reads coefficients from file on disk

        Coefficients are uniquely identified by their source/destination names

        """
        results_path = self._get_coefficients_path(source_name, destination_name)
        if os.path.isfile(results_path):
            return self._get_data_from_native_file(results_path)
        else:
            msg = "Could not find the coefficients file for %s to %s"
            self.logger.warning(msg, source_name, destination_name)
            return None

    def write_coefficients(self, source_name, destination_name, data):
        """Writes coefficients to file on disk

        Coefficients are uniquely identified by their source/destination names

        """
        results_path = self._get_coefficients_path(source_name, destination_name)
        self._write_data_to_native_file(results_path, data)

    def _get_coefficients_path(self, source_name, destination_name):

        results_dir = self.file_dir['coefficients']
        path = os.path.join(results_dir, source_name + '_' + destination_name)
        return path + '.dat'

    def read_results(self, modelrun_id, model_name, output_name, spatial_resolution,
                     temporal_resolution, timestep=None, modelset_iteration=None,
                     decision_iteration=None):
        """Return path to text file for a given output

        Parameters
        ----------
        modelrun_id : str
        model_name : str
        output_name : str
        spatial_resolution : str
        temporal_resolution : str
        timestep : int, optional
        modelset_iteration : int, optional
        decision_iteration : int, optional

        Returns
        -------
        data: numpy.ndarray

        """
        if timestep is None:
            raise NotImplementedError

        results_path = self._get_results_path(
            modelrun_id, model_name, output_name, spatial_resolution,
            temporal_resolution,
            timestep, modelset_iteration, decision_iteration)

        if self.storage_format == 'local_csv':
            csv_data = self._get_data_from_csv(results_path)
            region_names = self.read_region_names(spatial_resolution)
            interval_names = self.read_interval_names(temporal_resolution)
            return self.data_list_to_ndarray(csv_data, region_names, interval_names)
        elif self.storage_format == 'local_binary':
            return self._get_data_from_native_file(results_path)

    def write_results(self, modelrun_id, model_name, output_name, data, spatial_resolution,
                      temporal_resolution, timestep=None, modelset_iteration=None,
                      decision_iteration=None):
        """Return path to text file for a given output

        Parameters
        ----------
        modelrun_id : str
        model_name : str
        output_name : str
        data : numpy.ndarray
        spatial_resolution : str
        temporal_resolution : str
        timestep : int, optional
        modelset_iteration : int, optional
        decision_iteration : int, optional
        """
        if timestep is None:
            raise NotImplementedError

        results_path = self._get_results_path(
            modelrun_id, model_name, output_name, spatial_resolution,
            temporal_resolution,
            timestep, modelset_iteration, decision_iteration)
        os.makedirs(os.path.dirname(results_path), exist_ok=True)

        if data.ndim == 3:
            raise NotImplementedError
        elif data.ndim == 2:
            region_names = self.read_region_names(spatial_resolution)
            interval_names = self.read_interval_names(temporal_resolution)
            assert data.shape == (len(region_names), len(interval_names))

            if self.storage_format == 'local_csv':
                csv_data = self.ndarray_to_data_list(
                    data, region_names, interval_names, timestep=timestep)
                self._write_data_to_csv(results_path, csv_data)
            elif self.storage_format == 'local_binary':
                self._write_data_to_native_file(results_path, data)
        else:
            raise DataMismatchError(
                "Expected to write either timestep x region x interval or " +
                "region x interval data"
            )

    def results_exist(self, modelrun_name):
        """Checks whether modelrun results exists on the filesystem
        for a particular modelrun_name

        Parameters
        ----------
        modelrun_name: str

        Returns
        -------
        bool: True when results exist for this modelrun_name
        """
        previous_results_dir = os.path.join(self.file_dir['results'],
                                            modelrun_name)
        results = list(glob.iglob(os.path.join(previous_results_dir, '**/*.*'),
                                  recursive=True))

        return len(results) > 0

    def prepare_warm_start(self, modelrun_id):
        """Copy the results from the previous modelrun if available

        Parameters
        ----------
        modelrun_id: str

        Returns
        -------
        num: The timestep where the data store was recovered to
        """
        results_dir = os.path.join(self.file_dir['results'], modelrun_id)

        # Return if path to previous modelruns does not exist
        if not os.path.isdir(results_dir):
            self.logger.info("Warm start not possible because modelrun has "
                             "no previous results (path does not exist)")
            return None

        # Return if no results exist in last modelrun
        if not self.results_exist(modelrun_id):
            self.logger.info("Warm start not possible because the "
                             "modelrun does not have any results")
            return None

        # Return if previous results were stored in a different format
        previous_results_dir = os.path.join(self.file_dir['results'], modelrun_id)
        results = list(glob.iglob(os.path.join(previous_results_dir, '**/*.*'),
                                  recursive=True))
        for filename in results:
            if (
                    (self.storage_format == 'local_csv' and
                        not filename.endswith(".csv")) or
                    (self.storage_format == 'local_binary' and
                        not filename.endswith(".dat"))
                        ):
                self.logger.info("Warm start not possible because a different "
                                 "storage mode was used in the previous run")
                return None

        # Perform warm start
        self.logger.info("Warm start " + modelrun_id)

        # Get metadata for all results
        result_metadata = []
        for filename in glob.iglob(os.path.join(results_dir, '**/*.*'),
                                   recursive=True):
            result_metadata.append(self._parse_results_path(
                filename.replace(self.file_dir['results'], '')[1:]))

        # Find latest timestep
        result_metadata = sorted(result_metadata, key=lambda k: k['timestep'],
                                 reverse=True)
        latest_timestep = result_metadata[0]['timestep']

        # Remove all results with this timestep
        results_to_remove = \
            [result for result in result_metadata
             if result['timestep'] == latest_timestep]

        for result in results_to_remove:
            os.remove(self._get_results_path(result['modelrun_id'],
                      result['model_name'],
                      result['output_name'],
                      result['spatial_resolution'],
                      result['temporal_resolution'],
                      result['timestep'],
                      result['modelset_iteration'],
                      result['decision_iteration']))

        self.logger.info("Warm start will resume at timestep %s",
                         latest_timestep)
        return latest_timestep

    def _get_results_path(self, modelrun_id, model_name, output_name,
                          spatial_resolution,
                          temporal_resolution, timestep, modelset_iteration=None,
                          decision_iteration=None):
        """Return path to filename for a given output without file extension

        On the pattern of:
            results/
            <modelrun_name>/
            <model_name>/
            decision_<id>_modelset_<id>/ or decision_<id>/ or modelset_<id>/ or none
                output_<output_name>_
                timestep_<timestep>_
                regions_<spatial_resolution>_
                intervals_<temporal_resolution>.csv

        Parameters
        ----------
        modelrun_id : str
        model_name : str
        output_name : str
        spatial_resolution : str
        temporal_resolution : str
        timestep : str or int
        modelset_iteration : int, optional
        decision_iteration : int, optional

        Returns
        -------
        path : strs
        """
        results_dir = self.file_dir['results']
        if modelset_iteration is None and decision_iteration is None:
            path = os.path.join(
                results_dir,
                modelrun_id,
                model_name,
                "output_{}_timestep_{}_regions_{}_intervals_{}".format(
                    output_name,
                    timestep,
                    spatial_resolution,
                    temporal_resolution
                )
            )
        elif modelset_iteration is None and decision_iteration is not None:
            path = os.path.join(
                results_dir,
                modelrun_id,
                model_name,
                "decision_{}".format(decision_iteration),
                "output_{}_timestep_{}_regions_{}_intervals_{}".format(
                    output_name,
                    timestep,
                    spatial_resolution,
                    temporal_resolution
                )
            )
        elif modelset_iteration is not None and decision_iteration is None:
            path = os.path.join(
                results_dir,
                modelrun_id,
                model_name,
                "modelset_{}".format(modelset_iteration),
                "output_{}_timestep_{}_regions_{}_intervals_{}".format(
                    output_name,
                    timestep,
                    spatial_resolution,
                    temporal_resolution
                )
            )
        else:
            path = os.path.join(
                results_dir,
                modelrun_id,
                model_name,
                "decision_{}_modelset_{}".format(decision_iteration, modelset_iteration),
                "output_{}_timestep_{}_regions_{}_intervals_{}".format(
                    output_name,
                    timestep,
                    spatial_resolution,
                    temporal_resolution
                )
            )

        if self.storage_format == 'local_csv':
            path += '.csv'
        elif self.storage_format == 'local_binary':
            path += '.dat'

        return path

    def _parse_results_path(self, path):
        """Return result metadata for a given result path

        On the pattern of:
            results/
            <modelrun_name>/
            <model_name>/
            decision_<id>_modelset_<id>/ or decision_<id>/ or modelset_<id>/ or none
                output_<output_name>_
                timestep_<timestep>_
                regions_<spatial_resolution>_
                intervals_<temporal_resolution>.csv

        Parameters
        ----------
        path : str

        Returns
        -------
        dict : A dict containing all of the metadata
        """
        modelset_iteration = None
        decision_iteration = None

        data = re.findall(r"[\w']+", path)

        for section in data[2:len(data)]:
            if 'modelset' in section or 'decision' in section:
                regex_decision = re.findall(r"decision_(\d{1,})", section)
                regex_modelset = re.findall(r"modelset_(\d{1,})", section)
                if regex_decision:
                    decision_iteration = int(regex_decision[0])
                if regex_decision:
                    modelset_iteration = int(regex_modelset[0])
            elif section.startswith('output'):
                results = self._parse_output_section(section)
            elif section == 'csv':
                storage_format = 'local_csv'
            elif section == 'dat':
                storage_format = 'local_binary'

        return {
            'modelrun_id': data[0],
            'model_name': data[1],
            'output_name': '_'.join(results['output']),
            'spatial_resolution': '_'.join(results['regions']),
            'temporal_resolution': '_'.join(results['intervals']),
            'timestep': results['timestep'],
            'modelset_iteration': modelset_iteration,
            'decision_iteration': decision_iteration,
            'storage_format': storage_format
        }

    def _parse_output_section(self, section):
        result_elements = re.findall(r"[^_]+", section)
        results = {}
        parse_element = ""
        for element in result_elements:
            if element in ('output', 'timestep', 'regions', 'intervals') and \
                    parse_element != element:
                parse_element = element
            elif parse_element == 'output':
                results.setdefault('output', []).append(element)
            elif parse_element == 'timestep':
                results['timestep'] = int(element)
            elif parse_element == 'regions':
                results.setdefault('regions', []).append(element)
            elif parse_element == 'intervals':
                results.setdefault('intervals', []).append(element)
        return results

    def _read_project_config(self):
        """Read the project configuration

        Returns
        -------
        dict
            The project configuration
        """
        return self._read_yaml_file(self.file_dir['project'], 'project')

    def _write_project_config(self, data):
        """Write the project configuration

        Argument
        --------
        data: dict
            The project configuration
        """
        self._write_yaml_file(self.file_dir['project'], 'project', data)

    @staticmethod
    def _read_filenames_in_dir(path, extension):
        """Returns the name of the Yaml files in a certain directory

        Arguments
        ---------
        path: str
            Path to directory
        extension: str
            Extension of files (such as: '.yml' or '.csv')

        Returns
        -------
        list
            The list of files in `path` with extension
        """
        files = []
        for filename in os.listdir(path):
            if filename.endswith(extension):
                files.append(os.path.splitext(filename)[0])
        return files

    @staticmethod
    def _get_data_from_csv(filepath):
        scenario_data = []
        with open(filepath, 'r') as csvfile:
            reader = csv.DictReader(csvfile)
            scenario_data = []
            for row in reader:
                converted_row = {}
                
                converted_row['region'] = DatafileInterface._cast_str_to_int(row['region'])
                converted_row['interval'] = DatafileInterface._cast_str_to_int(row['interval'])
                
                if 'year' in row.keys():
                    converted_row['year'] = row['year']
                converted_row['value'] = row['value']
                
                scenario_data.append(converted_row)

        return scenario_data

    @staticmethod
    def _cast_str_to_int(value):
        if value.isdigit():
            return int(value)
        else:
            return value
                        

    @staticmethod
    def _write_data_to_csv(filepath, data):
        with open(filepath, 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=(
                'timestep',
                'region',
                'interval',
                'value'
            ))
            writer.writeheader()
            for row in data:
                writer.writerow(row)

    @staticmethod
    def _get_data_from_native_file(filepath):
        with pa.memory_map(filepath, 'rb') as f:
            f.seek(0)
            buf = f.read_buffer()

            data = pa.deserialize(buf)
        return data

    @staticmethod
    def _write_data_to_native_file(filepath, data):
        with pa.OSFile(filepath, 'wb') as f:
            f.write(
                pa.serialize(data).to_buffer()
            )

    @staticmethod
    def _read_yaml_file(path, filename, extension='.yml'):
        """Read a Data dict from a Yaml file

        Arguments
        ---------
        path: str
            Path to directory
        name: str
            Name of file
        extension: str, default='.yml'
            The file extension

        Returns
        -------
        dict
            The data of the Yaml file `filename` in `path`
        """
        filename = filename + extension
        filepath = os.path.join(path, filename)
        return load(filepath)

    @staticmethod
    def _write_yaml_file(path, filename, data, extension='.yml'):
        """Write a data dict to a Yaml file

        Arguments
        ---------
        path: str
            Path to directory
        name: str
            Name of file
        data: dict
            Data to be written to the file
        extension: str, default='.yml'
            The file extension
        """
        filename = filename + extension
        filepath = os.path.join(path, filename)
        dump(data, filepath)



"""Energy demand dummy model
"""
import numpy as np
from smif.model.sector_model import SectorModel


class EDMWrapper(SectorModel):
    """Energy model
    """
    def initialise(self, initial_conditions):
        pass

    def simulate(self, data):

        # Get the current timestep

        now = data.current_timestep
        self.logger.info("EDMWrapper received inputs in %s",
                         now)

        # Demonstrates how to get the value for a model parameter
        parameter_value = data.get_parameter('smart_meter_savings')
        self.logger.info('Smart meter savings: %s', parameter_value)

        # Demonstrates how to get the value for a model input
        # (defaults to the current time period)
        current_energy_demand = data.get_data('energy_demand')
        self.logger.info("Current energy demand in %s is %s",
                         now, current_energy_demand)

        # Demonstrates how to get the value for a model input from the base
        # timeperiod
        base_energy_demand = data.get_base_timestep_data('energy_demand')
        base_year = data.base_timestep
        self.logger.info("Base year energy demand in %s was %s", base_year,
                         base_energy_demand)

        # Demonstrates how to get the value for a model input from the previous
        # timeperiod
        if now > base_year:
            prev_energy_demand = data.get_previous_timestep_data('energy_demand')
            prev_year = data.previous_timestep
            self.logger.info("Previous energy demand in %s was %s",
                             prev_year, prev_energy_demand)

        # Pretend to call the 'energy model'
        # This code prints out debug logging messages for each input
        # defined in the energy_demand configuration
        for name in self.inputs.names:
            time_intervals = self.inputs[name].get_interval_names()
            regions = self.inputs[name].get_region_names()
            for i, region in enumerate(regions):
                for j, interval in enumerate(time_intervals):
                    self.logger.info(
                        "%s %s %s",
                        interval,
                        region,
                        data.get_data(name)[i, j])

        # Write pretend results to data handler
        data.set_results("cost", np.ones((3, 1)) * 3)
        data.set_results("water_demand", np.ones((3, 1)) * 3)

        self.logger.info("EDMWrapper produced outputs in %s",
                         now)

    def extract_obj(self, results):
        return 0



"""Test SectorModel and SectorModelBuilder
"""
from copy import copy
from unittest.mock import Mock

from pytest import fixture, raises
from smif.convert.area import get_register
from smif.metadata import Metadata, MetadataSet
from smif.model.sector_model import SectorModel, SectorModelBuilder
from smif.parameters import ParameterList


@fixture(scope='function')
def get_sector_model_config(setup_folder_structure, setup_runpy_file, setup_registers):

    path = setup_folder_structure
    water_supply_wrapper_path = str(
        path.join(
            'models', 'water_supply', '__init__.py'
        )
    )

    config = {
        "name": "water_supply",
        "description": 'a description',
        "path": water_supply_wrapper_path,
        "classname": "WaterSupplySectorModel",
        "inputs": [
            {
                'name': 'raininess',
                'spatial_resolution': 'LSOA',
                'temporal_resolution': 'annual',
                'units': 'milliliter'
            }
        ],
        "outputs": [
            {
                'name': 'cost',
                'spatial_resolution': 'LSOA',
                'temporal_resolution': 'annual',
                'units': 'million GBP'
            },
            {
                'name': 'water',
                'spatial_resolution': 'LSOA',
                'temporal_resolution': 'annual',
                'units': 'megaliter'
            }
        ],
        "initial_conditions": [],
        "interventions": [
            {"name": "water_asset_a", "location": "oxford"},
            {"name": "water_asset_b", "location": "oxford"},
            {"name": "water_asset_c", "location": "oxford"},
        ],
        "parameters": [
            {
                'name': 'assump_diff_floorarea_pp',
                'description': 'Difference in floor area per person \
                                in end year compared to base year',
                'absolute_range': (0.5, 2),
                'suggested_range': (0.5, 2),
                'default_value': 1,
                'units': '%'
            }
        ]
    }

    return config


class EmptySectorModel(SectorModel):

    def initialise(self, initial_conditions):
        pass

    def simulate(self, timestep, data=None):
        return {}

    def extract_obj(self, results):
        return 0


class TestCompositeSectorModel():

    def test_add_input(self):

        model = EmptySectorModel('test_model')
        model.add_input('input_name', [], [], 'units')

        inputs = model.inputs

        assert inputs.names == ['input_name']
        assert inputs.units == ['units']

        assert inputs['input_name'] == Metadata('input_name', [], [], 'units')

    def test_add_output(self):

        model = EmptySectorModel('test_model')
        model.add_output('output_name', Mock(), Mock(), 'units')

        outputs = model.outputs

        assert outputs.names == ['output_name']
        assert outputs.units == ['units']

    def test_run_sector_model(self):

        model = EmptySectorModel('test_model')
        model.add_input('input_name', [], [], 'units')
        data = {'input_name': [0]}
        actual = model.simulate(2010, data)
        assert actual == {}


class TestSectorModelBuilder():

    def test_add_inputs(self, setup_folder_structure, setup_runpy_file):

        model_path = str(setup_folder_structure.join('models', 'water_supply',
                                                     '__init__.py'))

        builder = SectorModelBuilder('test')
        builder.load_model(model_path, 'WaterSupplySectorModel')

        inputs = [{'name': 'an_input',
                   'spatial_resolution': 'LSOA',
                   'temporal_resolution': 'annual',
                   'units': 'tonnes'}]

        builder.add_inputs(inputs)

        assert 'an_input' in builder._sector_model.inputs.names

    def test_sector_model_builder(self, setup_folder_structure, setup_runpy_file):
        model_path = str(setup_folder_structure.join('models', 'water_supply',
                                                     '__init__.py'))

        register = Mock()
        register.get_entry = Mock(return_value='a_resolution_set')

        registers = {'regions': register,
                     'intervals': register}

        builder = SectorModelBuilder('water_supply', registers)
        builder.load_model(model_path, 'WaterSupplySectorModel')

        assets = [
            {
                'name': 'water_asset_a',
                'type': 'water_pump',
                'attributes': {
                    'capital_cost': 1000,
                    'economic_lifetime': 25,
                    'operational_lifetime': 25
                }
            }
        ]
        builder.add_interventions(assets)

        # builder.add_inputs(inputs)
        # builder.add_outputs(outputs)

        model = builder.finish()
        assert isinstance(model, SectorModel)

        assert model.name == 'water_supply'
        assert model.intervention_names == ['water_asset_a']
        assert model.interventions == assets

    def test_path_not_found(self):
        builder = SectorModelBuilder('water_supply', Mock())
        with raises(FileNotFoundError) as ex:
            builder.load_model('/fictional/path/to/model.py', 'WaterSupplySectorModel')
        msg = "Cannot find '/fictional/path/to/model.py' for the 'water_supply' model"
        assert msg in str(ex.value)

    def test_build_from_config(self, get_sector_model_config):
        config = get_sector_model_config
        builder = SectorModelBuilder('test_sector_model')
        timesteps = [2015, 2020]
        builder.construct(config, timesteps)
        sector_model = builder.finish()
        assert sector_model.name == 'water_supply'
        assert sector_model.timesteps == timesteps

        actual = sector_model.as_dict()
        # sort to match expected output
        actual['inputs'].sort(key=lambda m: m['name'])
        actual['outputs'].sort(key=lambda m: m['name'])
        assert actual == config


class TestInputs:

    def test_add_no_inputs(self, setup_folder_structure, setup_runpy_file):
        model_path = str(setup_folder_structure.join('models', 'water_supply', '__init__.py'))
        registers = {'regions': Mock(),
                     'intervals': Mock()}

        builder = SectorModelBuilder('water_supply_test', registers)
        builder.load_model(model_path, 'WaterSupplySectorModel')
        builder.add_inputs(None)
        sector_model = builder.finish()
        assert isinstance(sector_model.inputs, MetadataSet)
        actual_inputs = sector_model.inputs.names
        assert actual_inputs == []


class TestSectorModel(object):

    def test_interventions_names(self):
        assets = [
            {'name': 'water_asset_a'},
            {'name': 'water_asset_b'},
            {'name': 'water_asset_c'}
        ]
        model = EmptySectorModel('test_model')
        model.interventions = assets

        intervention_names = model.intervention_names

        assert len(intervention_names) == 3
        assert 'water_asset_a' in intervention_names
        assert 'water_asset_b' in intervention_names
        assert 'water_asset_c' in intervention_names

    def test_interventions(self):
        interventions = [
            {
                'name': 'water_asset_a',
                'capital_cost': 1000,
                'economic_lifetime': 25,
                'operational_lifetime': 25
            },
            {
                'name': 'water_asset_b',
                'capital_cost': 1500,
            },
            {
                'name': 'water_asset_c',
                'capital_cost': 3000,
            }
        ]
        model = EmptySectorModel('test_model')
        model.interventions = interventions
        actual = model.interventions

        assert actual == interventions

        assert sorted(model.intervention_names) == [
            'water_asset_a',
            'water_asset_b',
            'water_asset_c'
        ]


class TestParameters():

    def test_add_parameter(self):
        """Adding a parameter adds a reference to the parameter list entry to
        the model that contains it.
        """

        model = copy(EmptySectorModel('test_model'))
        model.simulate = lambda x, y: {'savings': y['smart_meter_savings']}

        param_config = {'name': 'smart_meter_savings',
                        'description': 'The savings from smart meters',
                        'absolute_range': (0, 100),
                        'suggested_range': (3, 10),
                        'default_value': 3,
                        'units': '%'}
        model.add_parameter(param_config)

        assert isinstance(model.parameters, ParameterList)
        assert model.parameters['smart_meter_savings'].as_dict() == param_config

        actual = model.simulate(2010, {'smart_meter_savings': 3})
        expected = {'savings': 3}
        assert actual == expected


class TestSectorModelIntervals():
    """SectorModels should have access to intervals (name)
    """

    def test_access_region_names(self):
        """Access names
        """
        model = EmptySectorModel('region_test')
        interval_names = model.get_interval_names('annual')
        assert interval_names == ['1']


class TestSectorModelRegions():
    """SectorModels should have access to regions (name, geometry and centroid)
    """
    def get_model(self):
        """Get a model with region register as setup in conftest
        """
        rreg = get_register()
        model = EmptySectorModel('region_test')
        model.regions = rreg
        return model

    def test_access_region_names(self):
        """Access names
        """
        model = self.get_model()
        region_names = model.get_region_names('half_squares')
        assert region_names == ['a', 'b']

    def test_access_region_geometries(self):
        model = self.get_model()
        actual = model.get_regions('half_squares')

        expected = [
            {
                'type': 'Feature',
                'properties': {'name': 'a'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': (((0.0, 0.0), (0.0, 1.0), (1.0, 1.0),
                                     (1.0, 0.0), (0.0, 0.0),),)
                }
            },
            {
                'type': 'Feature',
                'properties': {'name': 'b'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': (((0.0, 1.0), (0.0, 2.0), (1.0, 2.0),
                                     (1.0, 1.0), (0.0, 1.0),),)
                }
            },
        ]
        assert actual == expected

    def test_access_region_centroids(self):
        model = self.get_model()
        actual = model.get_region_centroids('half_squares')

        expected = [
            {
                'type': 'Feature',
                'properties': {'name': 'a'},
                'geometry': {
                    'type': 'Point',
                    'coordinates': (0.5, 0.5)
                }
            },
            {
                'type': 'Feature',
                'properties': {'name': 'b'},
                'geometry': {
                    'type': 'Point',
                    'coordinates': (0.5, 1.5)
                }
            },
        ]
        assert actual == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    conftest.py for smif.

    Read more about conftest.py under:
    https://pytest.org/latest/plugins.html
"""
from __future__ import absolute_import, division, print_function

import csv
import json
import logging
import os
from copy import copy

from pytest import fixture
from smif.convert.area import RegionSet
from smif.convert.area import get_register as get_region_register
from smif.convert.interval import IntervalSet
from smif.convert.interval import get_register as get_interval_register
from smif.convert.unit import get_register as get_unit_register
from smif.data_layer import DatafileInterface
from smif.data_layer.load import dump
from smif.parameters import Narrative

from .convert.conftest import (months, one_day, remap_months, remap_months_csv,
                               seasons, twenty_four_hours)
from .convert.test_area import (regions_half_squares, regions_half_triangles,
                                regions_rect, regions_single_half_square)

logging.basicConfig(filename='test_logs.log',
                    level=logging.DEBUG,
                    format='%(asctime)s %(name)-12s: %(levelname)-8s %(message)s',
                    filemode='w')


@fixture(scope='function')
def setup_folder_structure(tmpdir_factory, oxford_region, annual_intervals):
    """

    Returns
    -------
    :class:`LocalPath`
        Path to the temporary folder
    """
    folder_list = [
        'config',
        os.path.join('config', 'sos_model_runs'),
        os.path.join('config', 'sos_models'),
        os.path.join('config', 'sector_models'),
        'data',
        os.path.join('data', 'initial_conditions'),
        os.path.join('data', 'interval_definitions'),
        os.path.join('data', 'interventions'),
        os.path.join('data', 'narratives'),
        os.path.join('data', 'region_definitions'),
        os.path.join('data', 'scenarios'),
        os.path.join('data', 'coefficients'),
        os.path.join('data', 'strategies'),
        'models',
        'results'
    ]

    test_folder = tmpdir_factory.mktemp("smif")

    for folder in folder_list:
        test_folder.mkdir(folder)

    region_file = test_folder.join('data', 'region_definitions', 'test_region.json')
    region_file.write(json.dumps(oxford_region))

    intervals_file = test_folder.join('data', 'interval_definitions', 'annual.csv')
    intervals_file.write("id,start,end\n1,P0Y,P1Y\n")

    intervals_file = test_folder.join('data', 'interval_definitions', 'hourly.csv')
    intervals_file.write("id,start,end\n1,PT0H,PT1H\n")

    data = remap_months_csv()
    intervals_file = test_folder.join('data', 'interval_definitions', 'remap.csv')
    keys = data[0].keys()
    with open(str(intervals_file), 'w+') as open_csv_file:
        dict_writer = csv.DictWriter(open_csv_file, keys)
        dict_writer.writeheader()
        dict_writer.writerows(data)

    units_file = test_folder.join('data', 'user_units.txt')
    units_file.write("blobbiness = m^3 * 10^6\n")

    return test_folder


@fixture(scope='function')
def setup_runpy_file(setup_folder_structure):
    """The python script should contain an instance of SectorModel which wraps
    the sector model and allows it to be run.
    """
    base_folder = setup_folder_structure
    # Write a file for the water_supply model
    filename = base_folder.join('models', 'water_supply', '__init__.py')
    contents = """
from smif.model.sector_model import SectorModel

class WaterSupplySectorModel(SectorModel):
    def initialise(self, initial_conditions):
        pass

    def simulate(self, timestep, data=None):
        return {self.name: data}

    def extract_obj(self, results):
        return 0

"""
    filename.write(contents, ensure=True)


@fixture(scope='function')
def initial_system():
    """Initial system (interventions with build_date)
    """
    return [
        {
            'name': 'water_asset_a',
            'capacity': {
                'value': 5,
                'units': 'GW'
            },
            'operational_lifetime': {
                'value': 150,
                'units': "years"
            },
            'economic_lifetime': {
                'value': 50,
                'units': "years"
            },
            'capital_cost': {
                'value': 50,
                'units': "million £/km"
            },
            'location': 'oxford',
            'build_date': 2017
        },
        {
            'name': 'water_asset_b',
            'capacity': {
                'value': 15,
                'units': 'GW'
            },
            'operational_lifetime': {
                'value': 150,
                'units': "years"
            },
            'economic_lifetime': {
                'value': 50,
                'units': "years"
            },
            'capital_cost': {
                'value': 50,
                'units': "million £/km"
            },
            'location': 'oxford',
            'build_date': 2017
        },
        {
            'name': 'water_asset_c',
            'capacity': {
                'value': 25,
                'units': 'GW'
            },
            'operational_lifetime': {
                'value': 150,
                'units': "years"
            },
            'economic_lifetime': {
                'value': 50,
                'units': "years"
            },
            'capital_cost': {
                'value': 50,
                'units': "million £/km"
            },
            'location': 'oxford',
            'build_date': 2017
        }
    ]


@fixture(scope='function')
def initial_system_bis():
    """An extra intervention for the initial system
    """
    return [
        {
            'name': 'water_asset_d',
            'capacity': {
                'value': 15,
                'units': 'GW'
            },
            'operational_lifetime': {
                'value': 150,
                'units': "years"
            },
            'economic_lifetime': {
                'value': 50,
                'units': "years"
            },
            'capital_cost': {
                'value': 50,
                'units': "million £/km"
            },
            'location': 'oxford'
        }
    ]


@fixture
def parameters():
    return [
        {
            'name': 'smart_meter_savings',
            'description': 'The savings from smart meters',
            'absolute_range': (0, 100),
            'suggested_range': (3, 10),
            'default_value': 3,
            'units': '%'
        }
    ]


@fixture(scope='function')
def planned_interventions():
    """Return pre-specified planning intervention data
    """
    return [
        {
            'name': 'water_asset_a',
            'build_date': 2010,
            'attributes': [
                {
                    'key': 'new_capacity',
                    'value': 6
                },
                {
                    'key': 'description',
                    'value': 'Existing water treatment plants'
                },
                {
                    'key': 'location',
                    'value': {'lat': 51.74556, 'lon': -1.240528}
                }
            ]
        },
        {
            'name': 'water_asset_b',
            'build_date': 2010,
            'attributes': [
                {
                    'key': 'new_capacity',
                    'value': 6
                },
                {
                    'key': 'description',
                    'value': 'Existing water treatment plants'
                },
                {
                    'key': 'location',
                    'value': {'lat': 51.74556, 'lon': -1.240528}
                }
            ]
        },
        {
            'name': 'water_asset_c',
            'build_date': 2010,
            'attributes': [
                {
                    'key': 'new_capacity',
                    'value': 6
                },
                {
                    'key': 'description',
                    'value': 'Existing water treatment plants'
                },
                {
                    'key': 'location',
                    'value': {'lat': 51.74556, 'lon': -1.240528}
                }
            ]
        }
    ]


@fixture(scope='function')
def water_inputs():
    return [
        {
            'name': 'reservoir pumpiness',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual',
            'units': 'magnitude'
        }
    ]


@fixture(scope='function')
def water_outputs():
    return [
        {
            'name': 'storage_state',
            'spatial_resolution': 'national',
            'temporal_resolution': 'annual',
            'units': 'Ml'
        },
        {
            'name': 'storage_blobby',
            'spatial_resolution': 'national',
            'temporal_resolution': 'annual',
            'units': 'Ml'
        },
        {
            'name': 'total_water_demand',
            'spatial_resolution': 'national',
            'temporal_resolution': 'annual',
            'units': 'Ml'
        }
    ]


@fixture(scope='session')
def annual_intervals_csv():
    return [
        {
            "start": "P0Y",
            "end": "P1Y",
            "id": '1'
        }
    ]


@fixture(scope='session')
def annual_intervals():
    return [
        (
         '1', [("P0Y", "P1Y")]
        )
    ]


@fixture(scope='session')
def oxford_region():
    data = {
        "type": "FeatureCollection",
        "crs": {
            "type": "name",
            "properties": {
                "name": "urn:ogc:def:crs:EPSG::27700"
            }
        },
        "features": [
            {
                "type": "Feature",
                "properties": {
                    "name": "oxford"
                },
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [448180, 209366],
                            [449500, 211092],
                            [450537, 211029],
                            [450873, 210673],
                            [451250, 210793],
                            [451642, 210023],
                            [453855, 208466],
                            [454585, 208468],
                            [456077, 207967],
                            [456146, 207738],
                            [456668, 207779],
                            [456708, 207444],
                            [456278, 207122],
                            [456149, 206615],
                            [455707, 206798],
                            [455749, 204521],
                            [456773, 204488],
                            [457014, 204184],
                            [456031, 203475],
                            [456444, 202854],
                            [456087, 202044],
                            [455369, 201799],
                            [454396, 202203],
                            [453843, 201634],
                            [452499, 203209],
                            [452052, 203566],
                            [451653, 203513],
                            [450645, 205137],
                            [449497, 205548],
                            [449051, 206042],
                            [448141, 208446],
                            [448180, 209366]
                        ]
                    ]
                }
            },
        ]
    }
    return data


@fixture(scope='function')
def gb_ni_regions():
    data = {
        "type": "FeatureCollection",
        "crs": {
            "type": "name",
            "properties": {
                "name": "urn:ogc:def:crs:EPSG::27700"
            }
        },
        "features": [
            {
                "type": "Feature",
                "properties": {
                    "name": "GB"
                },
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [0, 1],
                            [0, 1.1],
                            [2, 3]
                        ]
                    ]
                }
            },
            {
                "type": "Feature",
                "properties": {
                    "name": "NI"
                },
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [2, 3],
                            [2, 3.2],
                            [4, 5]
                        ]
                    ]
                }
            },
        ]
    }
    return data


@fixture(scope='function')
def water_interventions_abc():
    return [
        {
            "name": "water_asset_a",
            "location": "oxford",
            "capital_cost": {
                "units": "£",
                "value": 1000
            },
            "economic_lifetime": {
                "units": "years",
                "value": 25
            },
            "operational_lifetime": {
                "units": "years",
                "value": 25
            }
        },
        {
            "name": "water_asset_b",
            "location": "oxford",
            "capital_cost": {
                "units": "£",
                "value": 1500
            },
            "economic_lifetime": {
                "units": "years",
                "value": 25
            },
            "operational_lifetime": {
                "units": "years",
                "value": 25
            }
        },
        {
            "name": "water_asset_c",
            "location": "oxford",
            "capital_cost": {
                "units": "£",
                "value": 3000
            },
            "economic_lifetime": {
                "units": "years",
                "value": 25
            },
            "operational_lifetime": {
                "units": "years",
                "value": 25
            }
        }
    ]


@fixture(scope="session", autouse=True)
def setup_registers(oxford_region, annual_intervals, tmpdir_factory):
    """One-time setup: load all the fixture region and interval
    sets into the module-level registers.
    """
    regions = get_region_register()
    lsoa = RegionSet('LSOA', oxford_region['features'])
    regions.register(lsoa)
    regions.register(regions_half_squares())
    regions.register(regions_single_half_square())
    regions.register(regions_half_triangles())
    regions.register(regions_rect())

    # register alt rect (same area)
    regions_rect_alt = copy(regions_rect())
    regions_rect_alt.name = 'rect_alt'
    regions.register(regions_rect_alt)

    intervals = get_interval_register()
    intervals.register(IntervalSet('annual', annual_intervals))
    intervals.register(IntervalSet('months', months()))
    intervals.register(IntervalSet('seasons', seasons()))
    intervals.register(IntervalSet('hourly_day', twenty_four_hours()))
    intervals.register(IntervalSet('one_day', one_day()))
    intervals.register(IntervalSet('remap_months', remap_months()))

    test_folder = tmpdir_factory.mktemp("smif")

    units_file = test_folder.join('user_units.txt')
    units_file.write("mcm = 10^6 * m^3\nGBP=[currency]\npeople=[people]\n")

    units = get_unit_register()
    units.register(str(units_file))


@fixture(scope='function')
def project_config():
    """Return sample project configuration
    """
    return {
        'project_name': 'NISMOD v2.0',
        'scenario_sets': [
            {
                'description': 'The annual change in UK population',
                'name': 'population',
                'facets': {'name': "population_count",
                           'description': "The count of population"}
            }
        ],
        'narrative_sets': [
            {
                'description': 'Defines the rate and nature of technological change',
                'name': 'technology'
            },
            {
                'description': 'Defines the nature of governance and influence upon decisions',
                'name': 'governance'
            }
        ],
        'region_definitions': [
            {
                'description': 'Local authority districts for the UK',
                'filename': 'test_region.json',
                'name': 'lad'
            }
        ],
        'interval_definitions': [
            {
                'description': 'The 8760 hours in the year named by hour',
                'filename': 'hourly.csv', 'name': 'hourly'
            },
            {
                'description': 'One annual timestep, used for aggregate yearly data',
                'filename': 'annual.csv', 'name': 'annual'
            },
            {
                'description': 'Remapped months to four representative months',
                'filename': 'remap.csv', 'name': 'remap_months'
            }
        ],
        'units': 'user_units.txt',
        'scenarios':
        [
            {
                'description': 'The High ONS Forecast for UK population out to 2050',
                'name': 'High Population (ONS)',
                'facets': [
                    {
                        'name': 'population_count',
                        'filename': 'population_high.csv',
                        'spatial_resolution': 'lad',
                        'temporal_resolution': 'annual',
                        'units': 'people',
                    }
                ],
                'scenario_set': 'population',
            },
            {
                'description': 'The Low ONS Forecast for UK population out to 2050',
                'name': 'Low Population (ONS)',
                'facets': [
                    {
                        'name': 'population_count',
                        'filename': 'population_low.csv',
                        'spatial_resolution': 'lad',
                        'temporal_resolution': 'annual',
                        'units': 'people',
                    }
                ],
                'scenario_set': 'population',
            }
        ],
        'narratives': [
            {
                'description': 'High penetration of SMART technology on the demand side',
                'filename': 'energy_demand_high_tech.yml',
                'name': 'Energy Demand - High Tech',
                'narrative_set': 'technology',
            },
            {
                'description': 'Stronger role for central government in planning and ' +
                               'regulation, less emphasis on market-based solutions',
                'filename': 'central_planning.yml',
                'name': 'Central Planning',
                'narrative_set': 'governance',
            }
        ]
    }


@fixture(scope='function')
def get_sos_model_run():
    """Return sample sos_model_run
    """
    return {
        'name': 'unique_model_run_name',
        'description': 'a description of what the model run contains',
        'stamp': '2017-09-20T12:53:23+00:00',
        'timesteps': [
            2015,
            2020,
            2025
        ],
        'sos_model': 'energy',
        'decision_module': 'energy_moea.py',
        'scenarios': {
            'population': 'High Population (ONS)'
        },
        'strategies': [{'strategy': 'pre-specified-planning',
                        'description': 'description of the strategy',
                        'model_name': 'energy_supply',
                        'interventions': [
                           {'name': 'nuclear_large', 'build_year': 2030},
                           {'name': 'carrington_return', 'build_year': 2030}
                           ]
                        }],
        'narratives': {
            'technology': [
                'Energy Demand - High Tech'
            ],
            'governance': [
                'Central Planning'
            ]
        }
    }


@fixture(scope='function')
def get_sos_model():
    """Return sample sos_model
    """
    return {
        'name': 'energy',
        'description': "A system of systems model which encapsulates "
                       "the future supply and demand of energy for the UK",
        'scenario_sets': [
            'population'
        ],
        'narrative_sets': [
            'technology'
        ],
        'sector_models': [
            'energy_demand',
            'energy_supply'
        ],
        'dependencies': [
            {
                'source_model': 'population',
                'source_model_output': 'count',
                'sink_model': 'energy_demand',
                'sink_model_input': 'population'
            },
            {
                'source_model': 'energy_demand',
                'source_model_output': 'gas_demand',
                'sink_model': 'energy_supply',
                'sink_model_input': 'natural_gas_demand'
            }
        ]
    }


@fixture(scope='function')
def get_sector_model():
    """Return sample sector_model
    """
    return {
        'name': 'energy_demand_sample',
        'description': "Computes the energy demand of the"
                       "UK population for each timestep",
        'classname': 'EnergyDemandWrapper',
        'path': '../../models/energy_demand/run.py',
        'inputs': [
            {
                'name': 'population',
                'spatial_resolution': 'lad',
                'temporal_resolution': 'annual',
                'units': 'people'
            }
        ],
        'outputs': [
            {
                'name': 'gas_demand',
                'spatial_resolution': 'lad',
                'temporal_resolution': 'hourly',
                'units': 'GWh'
            }
        ],
        'parameters': [
            {
                'absolute_range': '(0.5, 2)',
                'default_value': 1,
                'description': "Difference in floor area per person"
                               "in end year compared to base year",
                'name': 'assump_diff_floorarea_pp',
                'suggested_range': '(0.5, 2)',
                'units': 'percentage'
            }
        ],
        'interventions': ['energy_demand.yml'],
        'initial_conditions': ['energy_demand_init.yml']
    }


@fixture(scope='function')
def get_scenario_set():
    """Return sample scenario_set
    """
    return {
        "description": "Growth in UK economy",
        "name": "economy",
        "facets": {"name": "economy_low",
                   "description": "a description"}
    }


@fixture(scope='function')
def get_scenario():
    """Return sample scenario
    """
    return {
        "description": "Central Economy for the UK (High)",
        "name": "Central Economy (High)",
        "facets": [
            {
                "filename": "economy_low.csv",
                "name": "economy_low",
                "spatial_resolution": "national",
                "temporal_resolution": "annual",
                "units": "million people"
            }
        ],
        "scenario_set": "economy"
    }


@fixture(scope='function')
def get_narrative():
    """Return sample narrative
    """
    return {
        "description": "High penetration of SMART technology on the demand side",
        "filename": "high_tech_dsm.yml",
        "name": "High Tech Demand Side Management",
        "narrative_set": "technology"
    }


@fixture(scope='function')
def get_scenario_data():
    """Return sample scenario_data
    """
    return [
        {
            'value': 100,
            'units': 'people',
            'region': 'oxford',
            'interval': 1,
            'year': 2015
        },
        {
            'value': 150,
            'units': 'people',
            'region': 'oxford',
            'interval': 1,
            'year': 2016
        },
        {
            'value': 200,
            'units': 'people',
            'region': 'oxford',
            'interval': 1,
            'year': 2017
        }
    ]


@fixture(scope='function')
def narrative_data():
    """Return sample narrative_data
    """
    return {
        'energy_demand': {
            'smart_meter_savings': 8
        },
        'water_supply': {
            'clever_water_meter_savings': 8
        }
    }


@fixture(scope='function')
def get_handler(setup_folder_structure, project_config):
    basefolder = setup_folder_structure
    project_config_path = os.path.join(
        str(basefolder), 'config', 'project.yml')
    dump(project_config, project_config_path)

    return DatafileInterface(str(basefolder), 'local_binary')


@fixture
def get_narrative_obj():
    narrative = Narrative('Energy Demand - High Tech',
                          'A description',
                          'technology')
    return narrative


@fixture
def get_narrative_set():
    return {
        'name': 'technology',
        'description': 'Describes the evolution of technology'
    }



from flask import Flask
from smif.http_api.register import (register_api_endpoints,
                                    register_error_handlers, register_routes)


def create_app(static_folder, template_folder, data_interface, scheduler):
    """Create Flask app object
    """
    app = Flask(
        __name__,
        static_url_path='',
        static_folder=static_folder,
        template_folder=template_folder
    )

    app.config.data_interface = data_interface
    app.config.scheduler = scheduler

    register_routes(app)
    register_api_endpoints(app)
    register_error_handlers(app)

    return app



"""In this module, we implement the conversion across space and time

The :class:`SpaceTimeConvertor` is instantiated with data to convert,
and the names of the four source and destination spatio-temporal resolutions.

The :meth:`~SpaceTimeConvertor.convert` method returns a new
:class:`numpy.ndarray` for passing to a sector model.
"""
import logging

from smif.convert.area import get_register as get_region_register
from smif.convert.interval import get_register as get_interval_register
from smif.convert.unit import get_register as get_unit_register

__author__ = "Will Usher, Tom Russell, Roald Schoenmakers"
__copyright__ = "Will Usher, Tom Russell, Roald Schoenmakers"
__license__ = "mit"


class SpaceTimeUnitConvertor(object):
    """Handles the conversion of time and space for a list of values

    Notes
    -----
    Future development requires using a data object which allows multiple views
    upon the values across the three dimensions of time, space and units. This
    will then allow more efficient conversion across any one of these dimensions
    while holding the others constant.  One option could be
    :class:`collections.ChainMap`.

    """
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.regions = get_region_register()
        self.intervals = get_interval_register()
        self.units = get_unit_register()

    def convert(self, data,
                from_spatial, to_spatial,
                from_temporal, to_temporal,
                from_unit, to_unit):
        """Convert the data from set of regions and intervals to another

        Parameters
        ----------
        data: numpy.ndarray
            An array of values with dimensions regions x intervals
        from_spatial: str
            The name of the spatial resolution of the data
        to_spatial: str
            The name of the required spatial resolution
        from_temporal: str
            The name of the temporal resolution of the data
        to_temporal: str
            The name of the required temporal resolution
        from_unit: str
            The name of the unit of the data
        to_unit: str
            The name of the required unit

        Returns
        -------
        numpy.ndarray
            An array of data with dimensions regions x intervals
        """
        assert from_spatial in self.regions.names, \
            "Cannot convert from spatial resolution {}".format(from_spatial)
        assert to_spatial in self.regions.names, \
            "Cannot convert to spatial resolution {}".format(to_spatial)
        assert from_temporal in self.intervals.names, \
            "Cannot convert from temporal resolution {}".format(from_temporal)
        assert to_temporal in self.intervals.names, \
            "Cannot convert to temporal resolution {}".format(to_temporal)
        assert from_unit in self.units.names, \
            "Cannot convert from unit {}".format(from_unit)
        assert to_unit in self.units.names, \
            "Cannot convert to unit {}".format(to_unit)

        converted = data

        if from_spatial != to_spatial and from_temporal != to_temporal:

            results = self.regions.convert(converted,
                                           from_spatial,
                                           to_spatial)
            converted = self.intervals.convert(results,
                                               from_temporal,
                                               to_temporal)

        elif from_temporal != to_temporal:
            converted = self.intervals.convert(converted,
                                               from_temporal,
                                               to_temporal)

        elif from_spatial != to_spatial:
            converted = self.regions.convert(converted,
                                             from_spatial,
                                             to_spatial)

        converted = self.units.convert(converted,
                                       from_unit,
                                       to_unit)

        return converted



"""Database-backed data interface
"""

from smif.data_layer.data_interface import DataInterface


class DatabaseInterface(DataInterface):
    """ Read and write interface to Database
    """
    def __init__(self, config_path):
        raise NotImplementedError()

    def read_sos_model_runs(self):
        raise NotImplementedError()

    def write_sos_model_run(self, sos_model_run):
        raise NotImplementedError()

    def read_sos_models(self):
        raise NotImplementedError()

    def write_sos_model(self, sos_model):
        raise NotImplementedError()

    def read_sector_models(self):
        raise NotImplementedError()

    def read_sector_model(self, sector_model_name):
        raise NotImplementedError()

    def write_sector_model(self, sector_model):
        raise NotImplementedError()

    def read_units(self):
        raise NotImplementedError()

    def write_unit(self, unit):
        raise NotImplementedError()

    def read_regions(self):
        raise NotImplementedError()

    def read_region_data(self, region_name):
        raise NotImplementedError()

    def write_region(self, region):
        raise NotImplementedError()

    def read_intervals(self):
        raise NotImplementedError()

    def read_interval_data(self, interval_name):
        raise NotImplementedError()

    def write_interval(self, interval):
        raise NotImplementedError()

    def read_scenario_sets(self):
        raise NotImplementedError()

    def read_scenario_set(self, scenario_set_name):
        raise NotImplementedError()

    def read_scenario_data(self, scenario_name):
        raise NotImplementedError()

    def write_scenario_set(self, scenario_set):
        raise NotImplementedError()

    def write_scenario(self, scenario):
        raise NotImplementedError()

    def read_narrative_sets(self):
        raise NotImplementedError()

    def read_narrative_set(self, narrative_set_name):
        raise NotImplementedError()

    def read_narrative_data(self, narrative_name):
        raise NotImplementedError()

    def write_narrative_set(self, narrative_set):
        raise NotImplementedError()

    def write_narrative(self, narrative):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""Parse yaml config files, to construct sector models
"""
from ruamel.yaml import YAML


def load(file_path):
    """Parse yaml config file into plain data (lists, dicts and simple values)

    Parameters
    ----------
    file_path : str
        The path of the configuration file to parse
    """
    with open(file_path, 'r') as file_handle:
        return YAML().load(file_handle)


def dump(data, file_path):
    """Write plain data to a file as yaml

    Parameters
    ----------
    file_path : str
        The path of the configuration file to write
    data
        Data to write (should be lists, dicts and simple values)
    """
    with open(file_path, 'w') as file_handle:
        yaml = YAML()
        yaml.default_flow_style=False
        yaml.allow_unicode=True
        return yaml.dump(data, file_handle)
